# ***************************************************************************
# *   Copyright (c) 2015 Przemo Firszt <przemo@firszt.eu>                   *
# *   Copyright (c) 2016 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FemToolsCcx"
__author__ = "Przemo Firszt, Bernd Hahnebach"
__url__ = "https://www.freecad.org"

## \addtogroup FEM
#  @{

import os
import sys
import subprocess
import shutil

import FreeCAD

from femtools import femutils
from femtools import membertools

from PySide import QtCore  # there might be a special reason this is not guarded ?!?

if FreeCAD.GuiUp:
    from PySide import QtGui
    import FemGui


class FemToolsCcx(QtCore.QRunnable, QtCore.QObject):
    """

    Attributes
    ----------
    analysis : Fem::FemAnalysis
        FEM group analysis object
        has to be present, will be set in __init__
    solver : Fem::FemSolverObjectPython
        FEM solver object
        has to be present, will be set in __init__
    base_name : str
        name of .inp/.frd file (without extension)
        It is used to construct .inp file path that is passed to CalculiX ccx
    ccx_binary : str
    working_dir : str
    results_present : bool
        indicating if there are calculation results ready for us
    members : class femtools/membertools/AnalysisMember
        contains references to all analysis member except solvers and mesh
        Updated with update_objects
    """

    finished = QtCore.Signal(int)

    def __init__(self, analysis=None, solver=None, test_mode=False):
        """The constructor

        Parameters
        ----------
        analysis : Fem::FemAnalysis, optional
            analysis group as a container for all  objects needed for the analysis
        solver : Fem::FemSolverObjectPython, optional
            solver object to be used for this solve
        test_mode : bool, optional
            mainly used in unit tests
        """

        QtCore.QRunnable.__init__(self)
        QtCore.QObject.__init__(self)

        self.ccx_binary_present = False
        self.analysis = None
        self.solver = None

        # TODO if something will go wrong in __init__ do not continue,
        # but do not raise a exception, break in a smarter way

        if analysis:
            self.analysis = analysis
            if solver:
                # analysis and solver given
                self.solver = solver
            else:
                # analysis given, search for the solver
                self.find_solver()
                if not self.solver:
                    raise Exception("FEM: No solver found!")
        else:
            if solver:
                # solver given, search for the analysis
                self.solver = solver
                self.find_solver_analysis()
                if not self.analysis:
                    raise Exception(
                        "FEM: The solver was given as parameter, "
                        "but no analysis for this solver was found!"
                    )
            else:
                # neither analysis nor solver given, search both
                self.find_analysis()
                if not self.analysis:
                    raise Exception(
                        "FEM: No solver was given and either no active analysis "
                        "or no analysis at all or more than one analysis found!"
                    )
                self.find_solver()
                if not self.solver:
                    raise Exception("FEM: No solver found!")

        if self.analysis.Document is not self.solver.Document:
            raise Exception("FEM: The analysis and solver are not in the same document!")
        if self.solver not in self.analysis.Group:
            raise Exception("FEM: The solver is not part of the analysis Group!")

        # print(self.solver)
        # print(self.analysis)
        if self.analysis and self.solver:
            self.working_dir = ""
            self.ccx_binary = ""
            self.base_name = ""
            self.results_present = False
            if test_mode:
                self.test_mode = True
                self.ccx_binary_present = True
            else:
                self.test_mode = False
                self.ccx_binary_present = False
            self.result_object = None
        else:
            raise Exception(
                "FEM: Something went wrong, the exception should have been raised earlier!"
            )

    def purge_results(self):
        """Remove all result objects and result meshes from an analysis group"""
        from femresult.resulttools import purge_results as pr

        pr(self.analysis)

    def reset_mesh_purge_results_checked(self):
        """Reset mesh color, deformation and removes all result objects
        if preferences to keep them is not set.
        """
        self.fem_prefs = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Fem/General")
        keep_results_on_rerun = self.fem_prefs.GetBool("KeepResultsOnReRun", False)
        if not keep_results_on_rerun:
            # we remove the result objects only, not the postprocessing ones.
            # Reason: "Not keep results" means for the user override the data. For postprocessing
            #         this means keeping all filters, just change the data.
            from femresult.resulttools import purge_result_objects as purge

            purge(self.analysis)

    def reset_all(self):
        """Reset mesh color, deformation and removes all result objects"""
        self.purge_results()

    def _get_several_member(self, obj_type):
        return membertools.get_several_member(self.analysis, obj_type)

    def find_analysis(self):
        if FreeCAD.GuiUp:
            self.analysis = FemGui.getActiveAnalysis()
        if self.analysis:
            return
        found_analysis = False
        # search in the active document
        for m in FreeCAD.activeDocument().Objects:
            if femutils.is_of_type(m, "Fem::FemAnalysis"):
                if not found_analysis:
                    self.analysis = m
                    found_analysis = True
                else:
                    self.analysis = None  # more than one analysis
        if self.analysis:
            if FreeCAD.GuiUp:
                FemGui.setActiveAnalysis(self.analysis)

    def find_solver_analysis(self):
        """get the analysis group the solver belongs to"""
        if self.solver.getParentGroup():
            obj = self.solver.getParentGroup()
            if femutils.is_of_type(obj, "Fem::FemAnalysis"):
                self.analysis = obj
                if FreeCAD.GuiUp:
                    FemGui.setActiveAnalysis(self.analysis)

    def find_solver(self):
        found_solver_for_use = False
        for m in self.analysis.Group:
            if femutils.is_of_type(m, "Fem::SolverCcxTools"):
                # we are going to explicitly check for the ccx tools solver type only,
                # thus it is possible to have lots of framework solvers inside the analysis anyway
                # for some methods no solver is needed (purge_results) --> solver could be none
                # analysis has one solver and no solver was set --> use the one solver
                # analysis has more than one solver and no solver was set --> use solver none
                # analysis has no solver --> use solver none
                if not found_solver_for_use:
                    # no solver was found before
                    self.solver = m
                    found_solver_for_use = True
                else:
                    # another solver was found --> We have more than one solver
                    # we do not know which one to use, so we use none !
                    self.solver = None
                    FreeCAD.Console.PrintLog(
                        "FEM: More than one solver in the analysis "
                        "and no solver given to analyze. "
                        "No solver is set!\n"
                    )

    def update_objects(self):
        ## @var mesh
        #  mesh for the analysis
        self.mesh = None
        mesh, message = membertools.get_mesh_to_solve(self.analysis)
        if mesh is not None:
            self.mesh = mesh
        else:
            # the prerequisites will run anyway and they will print a message box anyway
            # thus do not print one here, but print a console warning
            FreeCAD.Console.PrintWarning(f"{message} The prerequisite check will fail.\n")

        ## @var members
        # members of the analysis. All except the solver and the mesh
        self.member = membertools.AnalysisMember(self.analysis)

    def check_prerequisites(self):
        FreeCAD.Console.PrintMessage("\n")  # because of time print in separate line
        FreeCAD.Console.PrintMessage("Check prerequisites...\n")
        message = ""
        # analysis
        if not self.analysis:
            message += "No active Analysis\n"
        # solver
        if not self.solver:
            message += "No solver object defined in the analysis\n"
        if not self.working_dir:
            message += "Working directory not set\n"
        if not os.path.isdir(self.working_dir):
            message += f"Working directory '{self.working_dir}' doesn't exist."
        from femtools.checksanalysis import check_member_for_solver_calculix

        message += check_member_for_solver_calculix(
            self.analysis, self.solver, self.mesh, self.member
        )
        return message

    def set_base_name(self, base_name=None):
        """
        Set base_name

        Parameters
        ----------
        base_name : str, optional
            base_name base name of .inp/.frd file (without extension).
            It is used to construct .inp file path that is passed to CalculiX ccx
        """
        if base_name is None:
            self.base_name = ""
        else:
            self.base_name = base_name
        # Update inp file name
        self.set_inp_file_name()

    def set_inp_file_name(self, inp_file_name=None):
        """
        Set inp file name. Normally inp file name is set by write_inp_file.
        That name is also used to determine location and name of frd result file.

        Parameters
        ----------
        inp_file_name : str, optional
            input file name path
        """
        if inp_file_name is not None:
            self.inp_file_name = inp_file_name
        else:
            self.inp_file_name = os.path.join(self.working_dir, (self.base_name + ".inp"))

    def setup_working_dir(self, param_working_dir=None, create=False):
        """Set working dir for solver execution.

        Parameters
        ----------
        param_working_dir :  str, optional
            directory to be used for writing
        create : bool, optional
            Should the working directory be created if it does not exist
        """
        self.working_dir = ""
        # try to use given working dir or overwrite with solver working dir
        fem_general_prefs = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Fem/General")
        if param_working_dir is not None:
            self.working_dir = param_working_dir
            if femutils.check_working_dir(self.working_dir) is not True:
                if create is True:
                    FreeCAD.Console.PrintMessage(
                        f"Dir given as parameter '{self.working_dir}' doesn't exist.\n"
                    )
                else:
                    FreeCAD.Console.PrintError(
                        "Dir given as parameter '{}' doesn't exist "
                        "and create parameter is set to False.\n".format(self.working_dir)
                    )
                    self.working_dir = femutils.get_pref_working_dir(self.solver)
                    FreeCAD.Console.PrintMessage(
                        f"Dir '{self.working_dir}' will be used instead.\n"
                    )
        elif fem_general_prefs.GetBool("OverwriteSolverWorkingDirectory", True) is False:
            self.working_dir = self.solver.WorkingDir
            if femutils.check_working_dir(self.working_dir) is not True:
                if self.working_dir == "":
                    FreeCAD.Console.PrintError(
                        "Working Dir is set to be used from solver object "
                        "but Dir from solver object '{}' is empty.\n".format(self.working_dir)
                    )
                else:
                    FreeCAD.Console.PrintError(
                        f"Dir from solver object '{self.working_dir}' doesn't exist.\n"
                    )
                self.working_dir = femutils.get_pref_working_dir(self.solver)
                FreeCAD.Console.PrintMessage(f"Dir '{self.working_dir}' will be used instead.\n")
        else:
            self.working_dir = femutils.get_pref_working_dir(self.solver)

        # check working_dir exist, if not use a tmp dir and inform the user
        if femutils.check_working_dir(self.working_dir) is not True:
            FreeCAD.Console.PrintError(
                f"Dir '{self.working_dir}' doesn't exist or cannot be created.\n"
            )
            self.working_dir = femutils.get_temp_dir(self.solver)
            FreeCAD.Console.PrintMessage(f"Dir '{self.working_dir}' will be used instead.\n")

        # Update inp file name
        self.set_inp_file_name()

    def write_inp_file(self):

        # get mesh set data
        # TODO use separate method for getting the mesh set data
        from femmesh import meshsetsgetter

        meshdatagetter = meshsetsgetter.MeshSetsGetter(
            self.analysis,
            self.solver,
            self.mesh,
            membertools.AnalysisMember(self.analysis),
        )
        # save the sets into the member objects of the instanz meshdatagetter
        meshdatagetter.get_mesh_sets()

        # write input file
        import femsolver.calculix.writer as iw

        self.inp_file_name = ""
        try:
            inp_writer = iw.FemInputWriterCcx(
                self.analysis,
                self.solver,
                self.mesh,
                meshdatagetter.member,
                self.working_dir,
                meshdatagetter.mat_geo_sets,
            )
            self.inp_file_name = inp_writer.write_solver_input()
        except Exception:
            FreeCAD.Console.PrintError(
                f"Unexpected error when writing CalculiX input file: {sys.exc_info()[1]}\n"
            )
            raise

    def setup_ccx(self, ccx_binary=None, ccx_binary_sig="CalculiX"):
        """Set Calculix binary path and validate its execution.

        Parameters
        ----------
        ccx_binary : str, optional
            It defaults to `None`. The path to the `ccx` binary. If it is `None`,
            the path is guessed.
        ccx_binary_sig : str, optional
            Defaults to 'CalculiX'. Expected output from `ccx` when run empty.

        """
        error_title = self.tr("No or wrong CalculiX binary ccx")
        self.ccx_binary = ccx_binary
        if self.ccx_binary is None:
            self.ccx_binary = FreeCAD.ParamGet(
                "User parameter:BaseApp/Preferences/Mod/Fem/Ccx"
            ).GetString("ccxBinaryPath", "")

        if not self.ccx_binary:
            # search in system
            self.ccx_binary = shutil.which("ccx")
        else:
            # check user defined path
            self.ccx_binary = shutil.which(self.ccx_binary)

        if self.ccx_binary is None:
            raise FileNotFoundError(
                "CalculiX binary not found\n"
                "Install CalculiX or set path to binary in FEM user preferences"
            )

        ccx_stdout = None
        ccx_stderr = None
        try:
            p = subprocess.Popen(
                [self.ccx_binary],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                shell=False,
                startupinfo=femutils.startProgramInfo(""),
            )
            ccx_stdout, ccx_stderr = p.communicate()
            if ccx_binary_sig in str(ccx_stdout):
                self.ccx_binary_present = True
            else:
                error_message = self.tr("FEM: wrong ccx binary")
                if FreeCAD.GuiUp:
                    QtGui.QMessageBox.critical(None, error_title, error_message)
                FreeCAD.Console.PrintError(error_message)
                # TODO: I'm still able to break it.
                # If user doesn't give a file but a path without a file or
                # a file which is not a binary no exception at all is raised.
        except OSError as e:
            FreeCAD.Console.PrintError(f"{e}\n")
            if e.errno == 2:
                error_message = self.tr(
                    "FEM: CalculiX binary ccx '{}' not found. "
                    "Please set the CalculiX binary ccx path in "
                    "FEM preferences tab CalculiX."
                ).format(self.ccx_binary)
                if FreeCAD.GuiUp:
                    QtGui.QMessageBox.critical(None, error_title, error_message)
                FreeCAD.Console.PrintError(error_message)

        except Exception as e:
            FreeCAD.Console.PrintError(f"{e}\n")
            error_message = self.tr(
                "FEM: CalculiX ccx '{}' output '{}' doesn't "
                "contain expected phrase '{}'. "
                "There are some problems when running the ccx binary. "
                "Check if ccx runs standalone without FreeCAD."
            ).format(self.ccx_binary, ccx_stdout, ccx_binary_sig)
            if FreeCAD.GuiUp:
                QtGui.QMessageBox.critical(None, error_title, error_message)
            FreeCAD.Console.PrintError(error_message)

    def start_ccx(self):
        import multiprocessing

        self.ccx_stdout = ""
        self.ccx_stderr = ""
        ont_backup = os.environ.get("OMP_NUM_THREADS")
        self.ccx_prefs = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Fem/Ccx")
        # If number of CPU's specified
        num_cpu_pref = self.ccx_prefs.GetInt("AnalysisNumCPUs", 1)
        if not ont_backup:
            ont_backup = str(num_cpu_pref)
        if num_cpu_pref > 1:
            # If user picked a number use that instead
            _env = os.putenv("OMP_NUM_THREADS", str(num_cpu_pref))
        else:
            _env = os.putenv("OMP_NUM_THREADS", str(multiprocessing.cpu_count()))
        # change cwd because ccx may crash if directory has no write permission
        # there is also a limit of the length of file names so jump to the document directory
        cwd = QtCore.QDir.currentPath()
        f = QtCore.QFileInfo(self.inp_file_name)
        QtCore.QDir.setCurrent(f.path())
        p = subprocess.Popen(
            [self.ccx_binary, "-i ", f.baseName()],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=False,
            env=_env,
        )
        self.ccx_stdout, self.ccx_stderr = p.communicate()
        self.ccx_stdout = self.ccx_stdout.decode()
        self.ccx_stderr = self.ccx_stderr.decode()
        os.putenv("OMP_NUM_THREADS", ont_backup)
        QtCore.QDir.setCurrent(cwd)
        return p.returncode

    def get_ccx_version(self):
        self.setup_ccx()
        import re
        from platform import system

        ccx_stdout = None
        ccx_stderr = None
        # Now extract the version number
        p = subprocess.Popen(
            [self.ccx_binary, "-v"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=False,
            startupinfo=femutils.startProgramInfo(""),
        )
        ccx_stdout, ccx_stderr = p.communicate()
        ccx_stdout = ccx_stdout.decode()
        m = re.search(r"(\d+).(\d+)", ccx_stdout)
        return (int(m.group(1)), int(m.group(2)))

    def ccx_run(self):
        ret_code = None
        FreeCAD.Console.PrintMessage("\n")  # because of time print in separate line
        FreeCAD.Console.PrintMessage("CalculiX solver run...\n")
        if self.test_mode:
            FreeCAD.Console.PrintError("CalculiX can not be run if test_mode is True.\n")
            return
        self.setup_ccx()
        if self.ccx_binary_present is False:
            error_message = (
                "FEM: CalculiX binary ccx '{}' not found. "
                "Please set the CalculiX binary ccx path in FEM preferences tab CalculiX.\n".format(
                    self.ccx_binary
                )
            )
            if FreeCAD.GuiUp:
                QtGui.QMessageBox.critical(None, "No CalculiX binary ccx", error_message)
            return
        progress_bar = FreeCAD.Base.ProgressIndicator()
        progress_bar.start("Everything seems fine. CalculiX ccx will be executed ...", 0)
        ret_code = self.start_ccx()
        self.finished.emit(ret_code)
        progress_bar.stop()
        if ret_code or self.ccx_stderr:
            if ret_code == 201 and self.solver.AnalysisType == "check":
                FreeCAD.Console.PrintMessage(
                    "It seems we run into NOANALYSIS problem, "
                    "thus workaround for wrong exit code for *NOANALYSIS check "
                    "and set ret_code to 0.\n"
                )
                # https://forum.freecad.org/viewtopic.php?f=18&t=31303&start=10#p260743
                ret_code = 0
            else:
                FreeCAD.Console.PrintError(f"CalculiX failed with exit code {ret_code}\n")
                FreeCAD.Console.PrintMessage("--------start of stderr-------\n")
                FreeCAD.Console.PrintMessage(self.ccx_stderr)
                FreeCAD.Console.PrintMessage("--------end of stderr---------\n")
                FreeCAD.Console.PrintMessage("--------start of stdout-------\n")
                FreeCAD.Console.PrintMessage(self.ccx_stdout)
                FreeCAD.Console.PrintMessage("\n--------end of stdout---------\n")
                FreeCAD.Console.PrintMessage("--------start problems---------\n")
                self.has_no_material_assigned()
                self.has_nonpositive_jacobians()
                FreeCAD.Console.PrintMessage("\n--------end problems---------\n")
        else:
            # remove highlighted nodes, if any
            if FreeCAD.GuiUp:
                self.mesh.ViewObject.HighlightedNodes = []

            FreeCAD.Console.PrintMessage("CalculiX finished without error.\n")
        return ret_code

    def run(self):
        self.update_objects()
        self.setup_working_dir()
        message = self.check_prerequisites()
        if message:
            text = "CalculiX can not be started due to missing prerequisites:\n"
            error_app = f"{text}{message}"
            error_gui = f"{text}\n{message}"
            FreeCAD.Console.PrintError(error_app)
            if FreeCAD.GuiUp:
                QtGui.QMessageBox.critical(None, "Missing prerequisite", error_gui)
            return False
        else:
            self.write_inp_file()
            if self.inp_file_name == "":
                error_message = "Error on writing CalculiX input file.\n"
                FreeCAD.Console.PrintError(error_message)
                if FreeCAD.GuiUp:
                    QtGui.QMessageBox.critical(None, "Error", error_message)
                return False
            else:
                FreeCAD.Console.PrintLog("Writing CalculiX input file completed.\n")
                ret_code = self.ccx_run()
                if ret_code is None:
                    error_message = "CalculiX has not been run. The CalculiX binary search returned: {}.\n".format(
                        self.ccx_binary_present
                    )
                    FreeCAD.Console.PrintError(error_message)
                    if FreeCAD.GuiUp:
                        QtGui.QMessageBox.critical(None, "Error", error_message)
                    return False
                if ret_code != 0:
                    error_message = f"CalculiX finished with error {ret_code}.\n"
                    FreeCAD.Console.PrintError(error_message)
                    if FreeCAD.GuiUp:
                        QtGui.QMessageBox.critical(None, "Error", error_message)
                    return False
                else:
                    FreeCAD.Console.PrintLog("Try to read result files\n")
                    self.load_results()
                    # TODO: output an error message if there where problems reading the results
        return True

    def has_no_material_assigned(self):
        if " *ERROR in calinput: no material was assigned" in self.ccx_stdout:
            without_material_elements = []
            without_material_elemnodes = []
            for line in self.ccx_stdout.splitlines():
                if "to element" in line:
                    # print(line)
                    # print(line.split())
                    non_mat_ele = int(line.split()[2])
                    # print(non_mat_ele)
                    if non_mat_ele not in without_material_elements:
                        without_material_elements.append(non_mat_ele)
            for e in without_material_elements:
                for n in self.mesh.FemMesh.getElementNodes(e):
                    without_material_elemnodes.append(n)
            without_material_elements = sorted(without_material_elements)
            without_material_elemnodes = sorted(without_material_elemnodes)
            command_for_withoutmatnodes = "without_material_elemnodes = {}".format(
                without_material_elemnodes
            )
            command_to_highlight = (
                "Gui.ActiveDocument.{}.HighlightedNodes = without_material_elemnodes".format(
                    self.mesh.Name
                )
            )
            # some output for the user
            FreeCAD.Console.PrintError(
                "\n\nCalculiX returned an error due to elements without materials.\n"
            )
            FreeCAD.Console.PrintMessage(
                f"without_material_elements = {without_material_elements}\n"
            )
            FreeCAD.Console.PrintMessage(command_for_withoutmatnodes + "\n")
            if FreeCAD.GuiUp:
                import FreeCADGui

                # with this the list without_material_elemnodes
                # will be available for further user interaction
                FreeCADGui.doCommand(command_for_withoutmatnodes)
                FreeCAD.Console.PrintMessage("\n")
                FreeCADGui.doCommand(command_to_highlight)
            FreeCAD.Console.PrintMessage(
                "\nFollowing some commands to copy. "
                "They will highlight the elements without materials "
                "or to reset the highlighted nodes:\n"
            )
            FreeCAD.Console.PrintMessage(command_to_highlight + "\n")
            # command to reset the Highlighted Nodes
            FreeCAD.Console.PrintMessage(
                f"Gui.ActiveDocument.{self.mesh.Name}.HighlightedNodes = []\n\n"
            )
            return True
        else:
            return False

    def has_nonpositive_jacobians(self):
        if "*ERROR in e_c3d: nonpositive jacobian" in self.ccx_stdout:
            nonpositive_jacobian_elements = []
            nonpositive_jacobian_elenodes = []
            for line in self.ccx_stdout.splitlines():
                if "determinant in element" in line:
                    # print(line)
                    # print(line.split())
                    non_posjac_ele = int(line.split()[3])
                    # print(non_posjac_ele)
                    if non_posjac_ele not in nonpositive_jacobian_elements:
                        nonpositive_jacobian_elements.append(non_posjac_ele)
            for e in nonpositive_jacobian_elements:
                for n in self.mesh.FemMesh.getElementNodes(e):
                    nonpositive_jacobian_elenodes.append(n)
            nonpositive_jacobian_elements = sorted(nonpositive_jacobian_elements)
            nonpositive_jacobian_elenodes = sorted(nonpositive_jacobian_elenodes)
            command_for_nonposjacnodes = "nonpositive_jacobian_elenodes = {}".format(
                nonpositive_jacobian_elenodes
            )
            command_to_highlight = (
                "Gui.ActiveDocument.{}.HighlightedNodes = nonpositive_jacobian_elenodes".format(
                    self.mesh.Name
                )
            )
            # some output for the user
            FreeCAD.Console.PrintError(
                "\n\nCalculiX returned an error due to nonpositive jacobian elements.\n"
            )
            FreeCAD.Console.PrintMessage(
                f"nonpositive_jacobian_elements = {nonpositive_jacobian_elements}\n"
            )
            FreeCAD.Console.PrintMessage(command_for_nonposjacnodes + "\n")
            if FreeCAD.GuiUp:
                import FreeCADGui

                # with this the list nonpositive_jacobian_elenodes
                # will be available for further user interaction
                FreeCADGui.doCommand(command_for_nonposjacnodes)
                FreeCAD.Console.PrintMessage("\n")
                FreeCADGui.doCommand(command_to_highlight)
            FreeCAD.Console.PrintMessage(
                "\nFollowing some commands to copy. "
                "They highlight the nonpositive jacobians "
                "or to reset the highlighted nodes:\n"
            )
            FreeCAD.Console.PrintMessage(command_to_highlight + "\n")
            # command to reset the Highlighted Nodes
            FreeCAD.Console.PrintMessage(
                f"Gui.ActiveDocument.{self.mesh.Name}.HighlightedNodes = []\n\n"
            )
            return True
        else:
            return False

    def load_results(self):
        FreeCAD.Console.PrintMessage("\n")  # because of time print in separate line
        FreeCAD.Console.PrintMessage("CalculiX read results...\n")
        self.results_present = False
        self.load_results_ccxfrd()
        self.load_results_ccxdat()
        self.analysis.Document.recompute()

    def load_results_ccxfrd(self):
        """Load results of ccx calculations from .frd file."""
        import feminout.importCcxFrdResults as importCcxFrdResults

        frd_result_file = os.path.splitext(self.inp_file_name)[0] + ".frd"
        if os.path.isfile(frd_result_file):
            importCcxFrdResults.importFrd(
                frd_result_file, self.analysis, "CCX_", self.solver.AnalysisType
            )
            for m in self.analysis.Group:
                if m.isDerivedFrom("Fem::FemResultObject"):
                    self.results_present = True
                    break
            else:
                if self.solver.AnalysisType == "check":
                    for m in self.analysis.Group:
                        if m.isDerivedFrom("Fem::FemMeshObjectPython"):
                            # we have no result object but a mesh object
                            # this happens in NOANALYSIS mode
                            break
                else:
                    FreeCAD.Console.PrintError("FEM: No result object in active Analysis.\n")
        else:
            FreeCAD.Console.PrintError(f"FEM: No frd result file found at {frd_result_file}\n")

    def load_results_ccxdat(self):
        """Load results of ccx calculations from .dat file."""
        import feminout.importCcxDatResults as importCcxDatResults

        dat_result_file = os.path.splitext(self.inp_file_name)[0] + ".dat"
        mode_frequencies = None
        dat_content = None

        if os.path.isfile(dat_result_file):
            mode_frequencies = importCcxDatResults.import_dat(dat_result_file, self.analysis)

            dat_file = open(dat_result_file)
            dat_content = dat_file.read()
            dat_file.close()
        else:
            FreeCAD.Console.PrintError(f"FEM: No dat result file found at {dat_result_file}\n")

        if mode_frequencies:
            # print(mode_frequencies)
            for m in self.analysis.Group:
                if m.isDerivedFrom("Fem::FemResultObject") and m.Eigenmode > 0:
                    for mf in mode_frequencies:
                        if m.Eigenmode == mf["eigenmode"]:
                            m.EigenmodeFrequency = mf["frequency"]

        if dat_content:
            # print(dat_content)
            dat_text_obj = self.analysis.Document.addObject("App::TextDocument", "ccx_dat_file")
            dat_text_obj.Text = dat_content
            dat_text_obj.setPropertyStatus("Text", "ReadOnly")  # set property editor readonly
            if FreeCAD.GuiUp:
                dat_text_obj.ViewObject.ReadOnly = True  # set editor view readonly
            self.analysis.addObject(dat_text_obj)


class CcxTools(FemToolsCcx):

    def __init__(self, solver=None):
        FemToolsCcx.__init__(self, None, solver)


##  @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2017 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""The BIM SetWPFront command"""

import FreeCAD
import FreeCADGui

QT_TRANSLATE_NOOP = FreeCAD.Qt.QT_TRANSLATE_NOOP


class BIM_SetWPFront:

    def GetResources(self):
        return {
            "Pixmap": "view-front.svg",
            "MenuText": QT_TRANSLATE_NOOP("BIM_SetWPFront", "Working Plane Front"),
            "ToolTip": QT_TRANSLATE_NOOP("BIM_SetWPFront", "Sets the working plane to Front"),
            "Accel": "W,P,1",
        }

    def Activated(self):
        FreeCADGui.addModule("WorkingPlane")
        FreeCADGui.doCommand("WorkingPlane.get_working_plane(update=False).set_to_front()")


class BIM_SetWPSide:

    def GetResources(self):
        return {
            "Pixmap": "view-right.svg",
            "MenuText": QT_TRANSLATE_NOOP("BIM_SetWPSide", "Working Plane Side"),
            "ToolTip": QT_TRANSLATE_NOOP("BIM_SetWPSide", "Sets the working plane to Side"),
            "Accel": "W,P,3",
        }

    def Activated(self):
        FreeCADGui.addModule("WorkingPlane")
        FreeCADGui.doCommand("WorkingPlane.get_working_plane(update=False).set_to_side()")


class BIM_SetWPTop:

    def GetResources(self):
        return {
            "Pixmap": "view-top.svg",
            "MenuText": QT_TRANSLATE_NOOP("BIM_SetWPTop", "Working Plane Top"),
            "ToolTip": QT_TRANSLATE_NOOP("BIM_SetWPTop", "Sets the working plane to Top"),
            "Accel": "W,P,2",
        }

    def Activated(self):
        FreeCADGui.addModule("WorkingPlane")
        FreeCADGui.doCommand("WorkingPlane.get_working_plane(update=False).set_to_top()")


class BIM_WPView:

    def GetResources(self):
        return {
            "Pixmap": "BIM_WPView",
            "MenuText": QT_TRANSLATE_NOOP("BIM_WPView", "Working Plane View"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "BIM_WPView",
                "Aligns the view to the current item in BIM Views window or to the current working plane",
            ),
            "Accel": "9",
        }

    def IsActive(self):
        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v

    def Activated(self):
        from bimcommands import BimViews
        import WorkingPlane

        vm = BimViews.findWidget()
        if vm:
            sel = vm.tree.selectedItems()
            if sel:
                # Aligning to current widget item
                BimViews.show(sel[0])
                return
            if hasattr(vm, "lastSelected"):
                # Aligning to stored widget item
                BimViews.show(vm.lastSelected)
                return
        # Aligning to current working plane
        WorkingPlane.get_working_plane().align_view()


FreeCADGui.addCommand("BIM_WPView", BIM_WPView())
FreeCADGui.addCommand("BIM_SetWPTop", BIM_SetWPTop())
FreeCADGui.addCommand("BIM_SetWPSide", BIM_SetWPSide())
FreeCADGui.addCommand("BIM_SetWPFront", BIM_SetWPFront())



# ***************************************************************************
# *   Copyright (c) 2005 Werner Mayer <werner.wm.mayer@gmx.de>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# *   Werner Mayer 2005                                                     *
# ***************************************************************************/

# Menu test module

import FreeCAD, os, unittest, FreeCADGui


# ---------------------------------------------------------------------------
# define the functions to test the FreeCAD base code
# ---------------------------------------------------------------------------

# def suite():
#    suite = unittest.TestSuite()
#    suite.addTest(DocTestCase("DocumentProperties"))
#    suite.addTest(DocTestCase("DocumentLabels"))
#    suite.addTest(DocTestCase("DocumentSaveAndRestore"))
#    return suite


class MenuCreateCases(unittest.TestCase):
    def setUp(self):
        FreeCADGui.activateWorkbench("TestWorkbench")
        FreeCAD.Console.PrintLog("Setup Test menu...\n")
        list = ["Test_TestAll", "Test_TestDoc", "Test_TestBase", "Test_TestWork"]
        w = FreeCADGui.getWorkbench("TestWorkbench")
        w.appendMenu("TestMenu", list)

    def testMenu(self):
        # check menu for items
        FreeCAD.Console.PrintLog("Checking Test menu...\n")
        w = FreeCADGui.getWorkbench("TestWorkbench")
        list = w.listMenus()
        self.b = False
        for i in list:
            if i == "TestMenu":
                self.b = True
        self.assertTrue(self.b, "Test menu not found")

    def tearDown(self):
        if self.b:
            FreeCAD.Console.PrintLog("Test menu successfully added\n")
        else:
            FreeCAD.Console.PrintLog("Adding Test menu failed\n")


class MenuDeleteCases(unittest.TestCase):
    def setUp(self):
        FreeCADGui.activateWorkbench("TestWorkbench")
        FreeCAD.Console.PrintLog("Remove Test menu...\n")
        w = FreeCADGui.getWorkbench("TestWorkbench")
        w.removeMenu("TestMenu")

    def testMenu(self):
        # check menu for items
        FreeCAD.Console.PrintLog("Checking Test menu...\n")
        w = FreeCADGui.getWorkbench("TestWorkbench")
        list = w.listMenus()
        self.b = True
        for i in list:
            if i == "TestMenu":
                self.b = False
        self.assertTrue(self.b == True, "Test menu still added")

    def tearDown(self):
        if self.b:
            FreeCAD.Console.PrintLog("Test menu successfully removed\n")
        else:
            FreeCAD.Console.PrintLog("Removing Test menu failed\n")



# ***************************************************************************
# *   Copyright (c) 2022 FreeCAD Project Association                        *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This library is free software; you can redistribute it and/or         *
# *   modify it under the terms of the GNU Lesser General Public            *
# *   License as published by the Free Software Foundation; either          *
# *   version 2.1 of the License, or (at your option) any later version.    *
# *                                                                         *
# *   This library is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU     *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with this library; if not, write to the Free Software   *
# *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA         *
# *   02110-1301  USA                                                       *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import unittest
import os
import codecs
import tempfile


class TestMetadata(unittest.TestCase):
    def setUp(self):
        self.test_dir = os.path.join(FreeCAD.getHomePath(), "Mod", "Test", "TestData")

    def test_xml_constructor(self):
        try:
            filename = os.path.join(self.test_dir, "basic_metadata.xml")
            md = FreeCAD.Metadata(filename)
        except Exception:
            self.fail("Metadata construction from XML file failed")

        with self.assertRaises(
            FreeCAD.Base.XMLBaseException,
            msg="Metadata construction from XML file with bad root node did not raise an exception",
        ):
            filename = os.path.join(self.test_dir, "bad_root_node.xml")
            md = FreeCAD.Metadata(filename)

        with self.assertRaises(
            FreeCAD.Base.XMLBaseException,
            msg="Metadata construction from invalid XML file did not raise an exception",
        ):
            filename = os.path.join(self.test_dir, "bad_xml.xml")
            md = FreeCAD.Metadata(filename)

        with self.assertRaises(
            FreeCAD.Base.XMLBaseException,
            msg="Metadata construction from XML file with invalid version did not raise an exception",
        ):
            filename = os.path.join(self.test_dir, "bad_version.xml")
            md = FreeCAD.Metadata(filename)

    def test_toplevel_tags(self):
        filename = os.path.join(self.test_dir, "basic_metadata.xml")
        md = FreeCAD.Metadata(filename)

        # Tags with only one element:
        self.assertEqual(md.Name, "Test Workbench")
        self.assertEqual(md.Description, "A package.xml file for unit testing.")
        self.assertEqual(md.Version, "1.0.1")
        self.assertEqual(md.Date, "2022-01-07")
        self.assertEqual(md.Icon, "Resources/icons/PackageIcon.svg")

        # Tags that are lists of elements:
        maintainers = md.Maintainer
        self.assertEqual(len(maintainers), 2)

        authors = md.Author
        self.assertEqual(len(authors), 3)

        urls = md.Urls
        self.assertEqual(len(urls), 5)

        tags = md.Tag
        self.assertEqual(len(tags), 2)

    def test_copy_constructor(self):
        filename = os.path.join(self.test_dir, "basic_metadata.xml")
        md = FreeCAD.Metadata(filename)
        copy_of_md = FreeCAD.Metadata(md)
        self.assertEqual(md.Name, copy_of_md.Name)
        self.assertEqual(md.Description, copy_of_md.Description)
        self.assertEqual(md.Version, copy_of_md.Version)

    def test_default_constructor(self):
        try:
            _ = FreeCAD.Metadata()
        except Exception:
            self.fail("Metadata default constructor failed")

    def test_content_types(self):
        filename = os.path.join(self.test_dir, "content_items.xml")
        md = FreeCAD.Metadata(filename)

        content = md.Content
        self.assertTrue("workbench" in content)
        self.assertTrue("preferencepack" in content)
        self.assertTrue("macro" in content)
        self.assertTrue("other_content_item" in content)

        workbenches = content["workbench"]
        preferencepacks = content["preferencepack"]
        macros = content["macro"]
        other = content["other_content_item"]

        self.assertEqual(len(workbenches), 4)
        self.assertEqual(len(macros), 2)
        self.assertEqual(len(preferencepacks), 1)

    def test_file_path(self):
        # Issue 7112
        try:
            filename = os.path.join(tempfile.gettempdir(), b"H\xc3\xa5vard.xml".decode("utf-8"))
            xmlfile = codecs.open(filename, mode="w", encoding="utf-8")
            xmlfile.write(
                r"""<?xml version="1.0" encoding="UTF-8" standalone="no" ?>
<package format="1" xmlns="https://wiki.freecad.org/Package_Metadata">
  <name>test</name>
  <description>Text</description>
  <version>1.0.0</version>
  <date>2022-01-01</date>
  <content>
    <workbench>
      <classname>Workbench</classname>
    </workbench>
  </content>
</package>"""
            )
            xmlfile.close()
            md = FreeCAD.Metadata(filename)
            self.assertEqual(md.Name, "test")
            self.assertEqual(md.Description, "Text")
            self.assertEqual(md.Version, "1.0.0")
        except UnicodeEncodeError as e:
            print("Ignore UnicodeEncodeError in test_file_path:\n{}".format(str(e)))

    def test_content_item_tags(self):
        filename = os.path.join(self.test_dir, "content_items.xml")
        md = FreeCAD.Metadata(filename)

        content = md.Content
        workbenches = content["workbench"]
        found = [False, False, False, False]
        for workbench in workbenches:
            print(workbench.Classname)
            if workbench.Classname == "TestWorkbenchA":
                found[0] = True
                self.assertEqual(workbench.FreeCADMin, "0.20.0")
                tags = workbench.Tag
                self.assertTrue("Tag A" in tags)
            elif workbench.Classname == "TestWorkbenchB":
                found[1] = True
                self.assertEqual(workbench.FreeCADMin, "0.1.0")
                self.assertEqual(workbench.FreeCADMax, "0.19.0")
                tags = workbench.Tag
                self.assertTrue("Tag B" in tags)
            elif workbench.Classname == "TestWorkbenchC":
                found[2] = True
                tags = workbench.Tag
                self.assertTrue("Tag C" in tags)
                self.assertTrue("Tag D" in tags)
            elif workbench.Classname == "TestWorkbenchD":
                found[3] = True
                dependencies = workbench.Depend
                expected_dependencies = {
                    "DependencyA": {"type": "automatic", "found": False},
                    "InternalWorkbench": {"type": "internal", "found": False},
                    "AddonWorkbench": {"type": "addon", "found": False},
                    "PythonPackage": {"type": "python", "found": False},
                    "DependencyB": {"type": "automatic", "found": False},
                }
                for dep in dependencies:
                    self.assertTrue(dep["package"] in expected_dependencies)
                    self.assertEqual(dep["type"], expected_dependencies[dep["package"]]["type"])
                    expected_dependencies[dep["package"]]["found"] = True
                for name, expected_dep in expected_dependencies.items():
                    self.assertTrue(expected_dep["found"], f"Failed to load dependency '{name}'")
        for f in found:
            self.assertTrue(
                f, f"One of the expected workbenches was not found in the metadata file"
            )

    def test_last_supported_version(self):
        pass

    def test_first_supported_version(self):
        pass

    def test_supports_current(self):
        pass

    def test_generic_metadata(self):
        pass

    def test_min_python_version(self):
        pass



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2025 Furgo                                              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

import Arch
from bimtests import TestArchBase


class TestArchPanel(TestArchBase.TestArchBase):

    def test_makePanel(self):
        """Test the makePanel function."""
        operation = "Testing makePanel function"
        self.printTestMessage(operation)

        panel = Arch.makePanel(name="TestPanel")
        self.assertIsNotNone(panel, "makePanel failed to create a panel object.")
        self.assertEqual(panel.Label, "TestPanel", "Panel label is incorrect.")

    # TODO: remove NOT_ prefix once it is understood why Arch.makePanelCut fails
    def NOT_test_makePanelCut(self):
        """Test the makePanelCut function."""
        operation = "Testing makePanelCut function"
        self.printTestMessage(operation)

        panel = Arch.makePanel(name="TestPanel")
        panel_cut = Arch.makePanelCut(panel, name="TestPanelCut")
        self.assertIsNotNone(panel_cut, "makePanelCut failed to create a panel cut object.")
        self.assertEqual(panel_cut.Label, "TestPanelCut", "Panel cut label is incorrect.")

    # TODO: remove NOT_ prefix once it is understood why Arch.makePanelSheet fails
    def NOT_test_makePanelSheet(self):
        """Test the makePanelSheet function."""
        operation = "Testing makePanelSheet function"
        self.printTestMessage(operation)

        panel_sheet = Arch.makePanelSheet(name="TestPanelSheet")
        self.assertIsNotNone(panel_sheet, "makePanelSheet failed to create a panel sheet object.")
        self.assertEqual(panel_sheet.Label, "TestPanelSheet", "Panel sheet label is incorrect.")



# ***************************************************************************
# *   Copyright (c) 2011 Juergen Riegel <FreeCAD@juergen-riegel.net>        *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

from math import pi, sqrt
import unittest

import FreeCAD


class TestPrimitive(unittest.TestCase):
    def setUp(self):
        self.Doc = FreeCAD.newDocument("PartDesignTestPrimitive")

    def testPrimitiveBox(self):
        self.Body = self.Doc.addObject("PartDesign::Body", "Body")
        self.Box = self.Doc.addObject("PartDesign::AdditiveBox", "Box")
        self.Box.Length = 11
        self.Box.Width = 11
        self.Box.Height = 11
        self.Body.addObject(self.Box)
        self.Doc.recompute()
        self.Box001 = self.Doc.addObject("PartDesign::SubtractiveBox", "Box001")
        self.Box001.Length = 10
        self.Box001.Width = 10
        self.Box001.Height = 10
        self.Body.addObject(self.Box001)
        self.Doc.recompute()
        self.assertAlmostEqual(self.Box001.Shape.Volume, 11**3 - 10**3)

    def testPrimitiveCylinder(self):
        self.Body = self.Doc.addObject("PartDesign::Body", "Body")
        self.Cylinder = self.Doc.addObject("PartDesign::AdditiveCylinder", "Cylinder")
        self.Cylinder.Radius = 11
        self.Cylinder.Height = 10
        self.Cylinder.Angle = 360
        self.Body.addObject(self.Cylinder)
        self.Doc.recompute()
        self.Cylinder001 = self.Doc.addObject("PartDesign::SubtractiveCylinder", "Cylinder001")
        self.Cylinder001.Radius = 10
        self.Cylinder001.Height = 10
        self.Cylinder001.Angle = 360
        self.Body.addObject(self.Cylinder001)
        self.Doc.recompute()
        self.assertAlmostEqual(self.Cylinder001.Shape.Volume, pi * 10 * (11**2 - 10**2))

    def testPrimitiveSphere(self):
        self.Body = self.Doc.addObject("PartDesign::Body", "Body")
        self.Sphere = self.Doc.addObject("PartDesign::AdditiveSphere", "Sphere")
        self.Sphere.Radius = 6
        self.Body.addObject(self.Sphere)
        self.Doc.recompute()
        self.Sphere001 = self.Doc.addObject("PartDesign::SubtractiveSphere", "Sphere001")
        self.Sphere001.Radius = 5
        self.Body.addObject(self.Sphere001)
        self.Doc.recompute()
        self.assertAlmostEqual(self.Sphere001.Shape.Volume, 4 / 3.0 * pi * (6**3 - 5**3))

    def testPrimitiveCone(self):
        self.Body = self.Doc.addObject("PartDesign::Body", "Body")
        self.Cone = self.Doc.addObject("PartDesign::AdditiveCone", "Cone")
        self.Cone.Radius1 = 0
        self.Cone.Radius2 = 4
        self.Cone.Height = 10
        self.Body.addObject(self.Cone)
        self.Doc.recompute()
        self.Cone001 = self.Doc.addObject("PartDesign::SubtractiveCone", "Cone")
        self.Cone001.Radius1 = 0
        self.Cone001.Radius2 = 3
        self.Cone001.Height = 10
        self.Body.addObject(self.Cone001)
        self.Doc.recompute()
        self.assertAlmostEqual(self.Cone001.Shape.Volume, 1 / 3.0 * pi * 10 * (4**2 - 3**2))

    def testPrimitiveEllipsoid(self):
        self.Body = self.Doc.addObject("PartDesign::Body", "Body")
        self.Ellipsoid = self.Doc.addObject("PartDesign::AdditiveEllipsoid", "Ellipsoid")
        self.Ellipsoid.Radius1 = 2
        self.Ellipsoid.Radius2 = 4
        self.Body.addObject(self.Ellipsoid)
        self.Doc.recompute()
        self.Ellipsoid001 = self.Doc.addObject("PartDesign::SubtractiveEllipsoid", "Ellipsoid001")
        self.Ellipsoid001.Radius1 = 1.5
        self.Ellipsoid001.Radius2 = 3
        self.Body.addObject(self.Ellipsoid001)
        self.Doc.recompute()
        self.assertAlmostEqual(
            self.Ellipsoid001.Shape.Volume, 4 / 3.0 * pi * (2 * 4**2 - 1.5 * 3**2), places=1
        )

    def testPrimitiveTorus(self):
        self.Body = self.Doc.addObject("PartDesign::Body", "Body")
        self.Torus = self.Doc.addObject("PartDesign::AdditiveTorus", "Torus")
        self.Torus.Radius1 = 10
        self.Torus.Radius2 = 4
        self.Body.addObject(self.Torus)
        self.Doc.recompute()
        self.Torus001 = self.Doc.addObject("PartDesign::SubtractiveTorus", "Torus001")
        self.Torus001.Radius1 = 10
        self.Torus001.Radius2 = 3
        self.Body.addObject(self.Torus001)
        self.Doc.recompute()
        self.assertAlmostEqual(self.Torus001.Shape.Volume, 2 * pi**2 * 10 * (4**2 - 3**2))

    def testPrimitivePrism(self):
        self.Body = self.Doc.addObject("PartDesign::Body", "Body")
        self.Prism = self.Doc.addObject("PartDesign::AdditivePrism", "Prism")
        self.Prism.Polygon = 6
        self.Prism.Circumradius = 4
        self.Prism.Height = 10
        self.Body.addObject(self.Prism)
        self.Doc.recompute()
        self.Prism001 = self.Doc.addObject("PartDesign::SubtractivePrism", "Prism001")
        self.Prism001.Polygon = 6
        self.Prism001.Circumradius = 3
        self.Prism001.Height = 10
        self.Body.addObject(self.Prism001)
        self.Doc.recompute()
        self.assertAlmostEqual(self.Prism001.Shape.Volume, 3 * sqrt(3) / 2.0 * 10 * (4**2 - 3**2))

    def testPrimitiveWedge(self):
        self.Body = self.Doc.addObject("PartDesign::Body", "Body")
        self.Wedge = self.Doc.addObject("PartDesign::AdditiveWedge", "Wedge")
        self.Wedge.X2min = 5
        self.Wedge.X2max = 5
        self.Wedge.Z2min = 0
        self.Wedge.Z2max = 10
        self.Body.addObject(self.Wedge)
        self.Doc.recompute()
        self.Wedge001 = self.Doc.addObject("PartDesign::SubtractiveWedge", "Wedge001")
        self.Wedge001.Xmin = 1
        self.Wedge001.Xmax = 9
        self.Wedge001.Ymax = 9
        self.Wedge001.X2min = 5
        self.Wedge001.X2max = 5
        self.Wedge001.Z2min = 0
        self.Wedge001.Z2max = 10
        self.Body.addObject(self.Wedge001)
        self.Doc.recompute()
        self.assertAlmostEqual(self.Wedge001.Shape.Volume, 1 / 2.0 * (10 * 10 - 9 * 8) * 10)

    def tearDown(self):
        # closing doc
        FreeCAD.closeDocument("PartDesignTestPrimitive")
        # print ("omit closing document for debugging")



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2015 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

__title__ = "Arch Material Management"
__author__ = "Yorik van Havre"
__url__ = "https://www.freecad.org"

## @package ArchMaterial
#  \ingroup ARCH
#  \brief The Material object and tools
#
#  This module provides tools to add materials to
#  Arch objects

import FreeCAD

from draftutils import params

if FreeCAD.GuiUp:
    import os
    from PySide import QtCore, QtGui
    from PySide.QtCore import QT_TRANSLATE_NOOP
    import FreeCADGui
    import Arch_rc  # Needed for access to icons # lgtm [py/unused_import]
    from draftutils.translate import translate
else:
    # \cond
    def translate(ctxt, txt):
        return txt

    def QT_TRANSLATE_NOOP(ctxt, txt):
        return txt

    # \endcond


class _ArchMaterialContainer:
    "The Material Container"

    def __init__(self, obj):
        self.Type = "MaterialContainer"
        obj.Proxy = self

    def execute(self, obj):
        return

    def dumps(self):
        if hasattr(self, "Type"):
            return self.Type

    def loads(self, state):
        if state:
            self.Type = state


class _ViewProviderArchMaterialContainer:
    "A View Provider for the Material Container"

    def __init__(self, vobj):
        vobj.Proxy = self

    def getIcon(self):
        return ":/icons/Arch_Material_Group.svg"

    def attach(self, vobj):
        self.Object = vobj.Object

    def setupContextMenu(self, vobj, menu):
        if FreeCADGui.activeWorkbench().name() != "BIMWorkbench":
            return
        actionMergeByName = QtGui.QAction(
            QtGui.QIcon(":/icons/Arch_Material_Group.svg"),
            translate("Arch", "Merge Duplicates"),
            menu,
        )
        actionMergeByName.triggered.connect(self.mergeByName)
        menu.addAction(actionMergeByName)

        actionReorder = QtGui.QAction(translate("Arch", "Reorder Children Alphabetically"), menu)
        actionReorder.triggered.connect(self.reorder)
        menu.addAction(actionReorder)

    def doubleClicked(self, vobj):
        """Handle double-click on the materials group in the Tree View.

        Open the BIM Materials dialog and return True to indicate the event was handled
        (prevents the tree from starting inline label editing, Qt's default behavior if the event
        is not handled).
        """
        FreeCADGui.runCommand("BIM_Material")
        return True

    def mergeByName(self):
        if hasattr(self, "Object"):
            mats = [o for o in self.Object.Group if o.isDerivedFrom("App::MaterialObject")]
            todelete = []
            for mat in mats:
                orig = None
                for om in mats:
                    if om.Label == mat.Label:
                        orig = om
                        break
                else:
                    if (
                        mat.Label[-1].isdigit()
                        and mat.Label[-2].isdigit()
                        and mat.Label[-3].isdigit()
                    ):
                        for om in mats:
                            if om.Label == mat.Label[:-3].strip():
                                orig = om
                                break
                if orig:
                    for par in mat.InList:
                        for prop in par.PropertiesList:
                            if getattr(par, prop) == mat:
                                FreeCAD.Console.PrintMessage(
                                    "Changed property '"
                                    + prop
                                    + "' of object "
                                    + par.Label
                                    + " from "
                                    + mat.Label
                                    + " to "
                                    + orig.Label
                                    + "\n"
                                )
                                setattr(par, prop, orig)
                    todelete.append(mat)
            for tod in todelete:
                if not tod.InList:
                    FreeCAD.Console.PrintMessage("Merging duplicate material " + tod.Label + "\n")
                    FreeCAD.ActiveDocument.removeObject(tod.Name)
                elif (len(tod.InList) == 1) and (
                    tod.InList[0].isDerivedFrom("App::DocumentObjectGroup")
                ):
                    FreeCAD.Console.PrintMessage("Merging duplicate material " + tod.Label + "\n")
                    FreeCAD.ActiveDocument.removeObject(tod.Name)
                else:
                    FreeCAD.Console.PrintMessage(
                        "Unable to delete material " + tod.Label + ": InList not empty\n"
                    )

    def reorder(self):
        if hasattr(self, "Object"):
            if hasattr(self.Object, "Group") and self.Object.Group:
                g = self.Object.Group
                g.sort(key=lambda obj: obj.Label)
                self.Object.Group = g
                FreeCAD.ActiveDocument.recompute()

    def dumps(self):
        return None

    def loads(self, state):
        return None


class _ArchMaterial:
    "The Material object"

    def __init__(self, obj):

        self.Type = "Material"
        obj.Proxy = self
        self.setProperties(obj)

    def onDocumentRestored(self, obj):

        self.setProperties(obj)

    def setProperties(self, obj):

        if not "Description" in obj.PropertiesList:
            obj.addProperty(
                "App::PropertyString",
                "Description",
                "Material",
                QT_TRANSLATE_NOOP("App::Property", "A description for this material"),
                locked=True,
            )
        if not "StandardCode" in obj.PropertiesList:
            obj.addProperty(
                "App::PropertyString",
                "StandardCode",
                "Material",
                QT_TRANSLATE_NOOP("App::Property", "A standard code (MasterFormat, OmniClass,…)"),
                locked=True,
            )
        if not "ProductURL" in obj.PropertiesList:
            obj.addProperty(
                "App::PropertyString",
                "ProductURL",
                "Material",
                QT_TRANSLATE_NOOP(
                    "App::Property", "A URL where to find information about this material"
                ),
                locked=True,
            )
        if not "Transparency" in obj.PropertiesList:
            obj.addProperty(
                "App::PropertyPercent",
                "Transparency",
                "Material",
                QT_TRANSLATE_NOOP("App::Property", "The transparency value of this material"),
                locked=True,
            )
        if not "Color" in obj.PropertiesList:
            obj.addProperty(
                "App::PropertyColor",
                "Color",
                "Material",
                QT_TRANSLATE_NOOP("App::Property", "The color of this material"),
                locked=True,
            )
        if not "SectionColor" in obj.PropertiesList:
            obj.addProperty(
                "App::PropertyColor",
                "SectionColor",
                "Material",
                QT_TRANSLATE_NOOP("App::Property", "The color of this material when cut"),
                locked=True,
            )

    def isSameColor(self, c1, c2):

        r = 4
        if round(c1[0], r) == round(c2[0], r):
            if round(c1[1], r) == round(c2[1], r):
                if round(c1[2], r) == round(c2[2], r):
                    return True
        return False

    def onChanged(self, obj, prop):

        d = obj.Material
        if prop == "Material":
            if "SectionColor" in obj.Material:
                c = tuple(
                    [
                        float(f)
                        for f in obj.Material["SectionColor"].strip("()").strip("[]").split(",")
                    ]
                )
                if hasattr(obj, "SectionColor"):
                    if not self.isSameColor(obj.SectionColor, c):
                        obj.SectionColor = c
            if "DiffuseColor" in obj.Material:
                c = tuple(
                    [
                        float(f)
                        for f in obj.Material["DiffuseColor"].strip("()").strip("[]").split(",")
                    ]
                )
                if hasattr(obj, "Color"):
                    if not self.isSameColor(obj.Color, c):
                        obj.Color = c
            if "Transparency" in obj.Material:
                t = int(obj.Material["Transparency"])
                if hasattr(obj, "Transparency"):
                    if obj.Transparency != t:
                        obj.Transparency = t
            if "ProductURL" in obj.Material:
                if hasattr(obj, "ProductURL"):
                    if obj.ProductURL != obj.Material["ProductURL"]:
                        obj.ProductURL = obj.Material["ProductURL"]
            if "StandardCode" in obj.Material:
                if hasattr(obj, "StandardCode"):
                    if obj.StandardCode != obj.Material["StandardCode"]:
                        obj.StandardCode = obj.Material["StandardCode"]
            if "Description" in obj.Material:
                if hasattr(obj, "Description"):
                    if obj.Description != obj.Material["Description"]:
                        obj.Description = obj.Material["Description"]
            if "Name" in obj.Material:
                if hasattr(obj, "Label"):
                    if obj.Label != obj.Material["Name"]:
                        obj.Label = obj.Material["Name"]
        elif prop == "Label":
            if "Name" in d:
                if d["Name"] == obj.Label:
                    return
            d["Name"] = obj.Label
        elif prop == "SectionColor":
            if hasattr(obj, "SectionColor"):
                if "SectionColor" in d:
                    if self.isSameColor(
                        tuple(
                            [float(f) for f in d["SectionColor"].strip("()").strip("[]").split(",")]
                        ),
                        obj.SectionColor[:3],
                    ):
                        return
                d["SectionColor"] = str(obj.SectionColor[:3])
        elif prop == "Color":
            if hasattr(obj, "Color"):
                if "DiffuseColor" in d:
                    if self.isSameColor(
                        tuple(
                            [float(f) for f in d["DiffuseColor"].strip("()").strip("[]").split(",")]
                        ),
                        obj.Color[:3],
                    ):
                        return
                d["DiffuseColor"] = str(obj.Color[:3])
        elif prop == "Transparency":
            if hasattr(obj, "Transparency"):
                val = str(obj.Transparency)
                if "Transparency" in d:
                    if d["Transparency"] == val:
                        return
                d["Transparency"] = val
        elif prop == "ProductURL":
            if hasattr(obj, "ProductURL"):
                val = obj.ProductURL
                if "ProductURL" in d:
                    if d["ProductURL"] == val:
                        return
                obj.Material["ProductURL"] = val
        elif prop == "StandardCode":
            if hasattr(obj, "StandardCode"):
                val = obj.StandardCode
                if "StandardCode" in d:
                    if d["StandardCode"] == val:
                        return
                d["StandardCode"] = val
        elif prop == "Description":
            if hasattr(obj, "Description"):
                val = obj.Description
                if "Description" in d:
                    if d["Description"] == val:
                        return
                d["Description"] = val
        if d and (d != obj.Material):
            obj.Material = d
            # if FreeCAD.GuiUp:
            # import FreeCADGui
            # not sure why this is needed, but it is...
            # FreeCADGui.ActiveDocument.resetEdit()

    def execute(self, obj):
        if obj.Material:
            if FreeCAD.GuiUp:
                c = None
                t = None
                if "DiffuseColor" in obj.Material:
                    c = tuple(
                        [
                            float(f)
                            for f in obj.Material["DiffuseColor"].strip("()").strip("[]").split(",")
                        ]
                    )
                if "Transparency" in obj.Material:
                    t = int(obj.Material["Transparency"])
                for p in obj.InList:
                    if (
                        hasattr(p, "Material")
                        and p.Material.Name == obj.Name
                        and getattr(obj.ViewObject, "UseMaterialColor", True)
                    ):
                        if c:
                            p.ViewObject.ShapeColor = c
                        if t:
                            p.ViewObject.Transparency = t
        return

    def dumps(self):
        if hasattr(self, "Type"):
            return self.Type

    def loads(self, state):
        if state:
            self.Type = state


class _ViewProviderArchMaterial:
    "A View Provider for the Material object"

    def __init__(self, vobj):
        vobj.Proxy = self

    def getIcon(self):
        if hasattr(self, "icondata"):
            return self.icondata
        return ":/icons/Arch_Material.svg"

    def attach(self, vobj):
        self.Object = vobj.Object

    def updateData(self, obj, prop):
        if prop == "Color":
            from PySide import QtCore, QtGui

            # custom icon
            if hasattr(obj, "Color"):
                c = obj.Color
                matcolor = QtGui.QColor(int(c[0] * 255), int(c[1] * 255), int(c[2] * 255))
                darkcolor = QtGui.QColor(int(c[0] * 125), int(c[1] * 125), int(c[2] * 125))
                im = QtGui.QImage(48, 48, QtGui.QImage.Format_ARGB32)
                im.fill(QtCore.Qt.transparent)
                pt = QtGui.QPainter(im)
                pt.setPen(QtGui.QPen(QtCore.Qt.black, 2, QtCore.Qt.SolidLine, QtCore.Qt.FlatCap))
                # pt.setBrush(QtGui.QBrush(matcolor, QtCore.Qt.SolidPattern))
                gradient = QtGui.QLinearGradient(0, 0, 48, 48)
                gradient.setColorAt(0, matcolor)
                gradient.setColorAt(1, darkcolor)
                pt.setBrush(QtGui.QBrush(gradient))
                pt.drawEllipse(6, 6, 36, 36)
                pt.setPen(QtGui.QPen(QtCore.Qt.white, 1, QtCore.Qt.SolidLine, QtCore.Qt.FlatCap))
                pt.setBrush(QtGui.QBrush(QtCore.Qt.white, QtCore.Qt.SolidPattern))
                pt.drawEllipse(12, 12, 12, 12)
                pt.end()

                ba = QtCore.QByteArray()
                b = QtCore.QBuffer(ba)
                b.open(QtCore.QIODevice.WriteOnly)
                im.save(b, "XPM")
                self.icondata = ba.data().decode("latin1")
                obj.ViewObject.signalChangeIcon()

    def onChanged(self, vobj, prop):
        if prop == "Material":
            if "Father" in vobj.Object.Material:
                for o in FreeCAD.ActiveDocument.Objects:
                    if o.isDerivedFrom("App::MaterialObject"):
                        if o.Label == vobj.Object.Material["Father"]:
                            o.touch()

    def setEdit(self, vobj, mode):
        if mode != 0:
            return None

        self.taskd = _ArchMaterialTaskPanel(vobj.Object)
        FreeCADGui.Control.showDialog(self.taskd)
        self.taskd.form.FieldName.setFocus()
        self.taskd.form.FieldName.selectAll()
        return True

    def unsetEdit(self, vobj, mode):
        if mode != 0:
            return None

        FreeCADGui.Control.closeDialog()
        return True

    def setupContextMenu(self, vobj, menu):
        if FreeCADGui.activeWorkbench().name() != "BIMWorkbench":
            return
        actionEdit = QtGui.QAction(translate("Arch", "Edit"), menu)
        actionEdit.triggered.connect(self.edit)
        menu.addAction(actionEdit)

    def edit(self):
        FreeCADGui.ActiveDocument.setEdit(self.Object, 0)

    def setTaskValue(self, widgetname, value):
        if hasattr(self, "taskd"):
            if hasattr(self.taskd, "form"):
                if hasattr(self.taskd.form, widgetname):
                    widget = getattr(self.taskd.form, widgetname)
                    if hasattr(widget, "setText"):
                        widget.setText(value)
                    elif hasattr(widget, "setValue"):
                        widget.setValue(value)

    def dumps(self):
        return None

    def loads(self, state):
        return None

    def claimChildren(self):
        ch = []
        if hasattr(self, "Object"):
            for o in self.Object.Document.Objects:
                if o.isDerivedFrom("App::MaterialObject"):
                    if o.Material:
                        if "Father" in o.Material:
                            if o.Material["Father"] == self.Object.Label:
                                ch.append(o)
        return ch


class _ArchMaterialTaskPanel:
    """The editmode TaskPanel for Arch Material objects"""

    def __init__(self, obj=None):
        self.cards = None
        self.existingmaterials = []
        self.obj = obj
        self.form = FreeCADGui.PySideUic.loadUi(":/ui/ArchMaterial.ui")
        colorPix = QtGui.QPixmap(16, 16)
        colorPix.fill(QtGui.QColor(204, 204, 204))
        self.form.ButtonColor.setIcon(QtGui.QIcon(colorPix))
        self.form.ButtonSectionColor.setIcon(QtGui.QIcon(colorPix))
        self.form.ButtonUrl.setIcon(QtGui.QIcon(":/icons/internet-web-browser.svg"))
        self.form.comboBox_MaterialsInDir.currentIndexChanged.connect(self.chooseMat)
        self.form.comboBox_FromExisting.currentIndexChanged.connect(self.fromExisting)
        self.form.comboFather.currentTextChanged.connect(self.setFather)
        self.form.ButtonColor.pressed.connect(self.getColor)
        self.form.ButtonSectionColor.pressed.connect(self.getSectionColor)
        self.form.ButtonUrl.pressed.connect(self.openUrl)
        self.form.ButtonEditor.pressed.connect(self.openEditor)
        self.form.ButtonCode.pressed.connect(self.getCode)
        self.fillMaterialCombo()
        self.fillExistingCombo()
        try:
            from bimcommands import BimClassification
        except Exception:
            self.form.ButtonCode.hide()
        else:
            self.form.ButtonCode.setIcon(QtGui.QIcon(":/icons/BIM_Classification.svg"))
        if self.obj:
            if hasattr(self.obj, "Material"):
                self.material = self.obj.Material
        self.setFields()

    def setFields(self):
        "sets the task box contents from self.material"
        if "Name" in self.material:
            self.form.FieldName.setText(self.material["Name"])
        elif self.obj:
            self.form.FieldName.setText(self.obj.Label)
        if "Description" in self.material:
            self.form.FieldDescription.setText(self.material["Description"])
        if "DiffuseColor" in self.material:
            self.form.ButtonColor.setIcon(self.getColorIcon(self.material["DiffuseColor"]))
        elif "ViewColor" in self.material:
            self.form.ButtonColor.setIcon(self.getColorIcon(self.material["ViewColor"]))
        elif "Color" in self.material:
            self.form.ButtonColor.setIcon(self.getColorIcon(self.material["Color"]))
        if "SectionColor" in self.material:
            self.form.ButtonSectionColor.setIcon(self.getColorIcon(self.material["SectionColor"]))
        if "StandardCode" in self.material:
            self.form.FieldCode.setText(self.material["StandardCode"])
        if "ProductURL" in self.material:
            self.form.FieldUrl.setText(self.material["ProductURL"])
        if "Transparency" in self.material:
            self.form.SpinBox_Transparency.setValue(int(self.material["Transparency"]))
        if "Father" in self.material:
            father = self.material["Father"]
        else:
            father = None
        found = False
        self.form.comboFather.addItem("None")
        for o in FreeCAD.ActiveDocument.Objects:
            if o.isDerivedFrom("App::MaterialObject"):
                if o != self.obj:
                    self.form.comboFather.addItem(o.Label)
                    if o.Label == father:
                        self.form.comboFather.setCurrentIndex(self.form.comboFather.count() - 1)
                        found = True
        if father and not found:
            self.form.comboFather.addItem(father)
            self.form.comboFather.setCurrentIndex(self.form.comboFather.count() - 1)

    def getColorIcon(self, color):
        if color:
            if "(" in color:
                c = tuple([float(f) for f in color.strip("()").split(",")])
                qcolor = QtGui.QColor()
                qcolor.setRgbF(c[0], c[1], c[2])
                colorPix = QtGui.QPixmap(16, 16)
                colorPix.fill(qcolor)
                icon = QtGui.QIcon(colorPix)
                return icon
        return QtGui.QIcon()

    def getFields(self):
        "sets self.material from the contents of the task box"
        self.material["Name"] = self.form.FieldName.text()
        self.material["Description"] = self.form.FieldDescription.text()
        self.material["DiffuseColor"] = self.getColorFromIcon(self.form.ButtonColor.icon())
        self.material["ViewColor"] = self.material["DiffuseColor"]
        self.material["Color"] = self.material["DiffuseColor"]
        self.material["SectionColor"] = self.getColorFromIcon(self.form.ButtonSectionColor.icon())
        self.material["StandardCode"] = self.form.FieldCode.text()
        self.material["ProductURL"] = self.form.FieldUrl.text()
        self.material["Transparency"] = str(self.form.SpinBox_Transparency.value())

    def getColorFromIcon(self, icon):
        "gets pixel color from the given icon"
        pixel = icon.pixmap(16, 16).toImage().pixel(0, 0)
        return str(QtGui.QColor(pixel).getRgbF())

    def accept(self):
        self.getFields()
        if self.obj:
            if hasattr(self.obj, "Material"):
                self.obj.Material = self.material
                self.obj.Label = self.material["Name"]
        FreeCAD.ActiveDocument.recompute()
        FreeCADGui.ActiveDocument.resetEdit()
        return True

    def reject(self):
        FreeCADGui.ActiveDocument.resetEdit()
        return True

    def chooseMat(self, card):
        "sets self.material from a card"
        card = self.form.comboBox_MaterialsInDir.currentText()
        if card in self.cards:
            import importFCMat

            self.material = importFCMat.read(self.cards[card])
            self.setFields()

    def fromExisting(self, index):
        "sets the contents from an existing material"
        if index > 0:
            if index <= len(self.existingmaterials):
                m = self.existingmaterials[index - 1]
                if m.Material:
                    self.material = m.Material
                    self.setFields()

    def setFather(self, text):
        "sets the father"
        if text:
            if text == "None":
                if "Father" in self.material:
                    # for some have Father at first and change to none
                    self.material.pop("Father")
            else:
                self.material["Father"] = text

    def getColor(self):
        self.getColorForButton(self.form.ButtonColor)

    def getSectionColor(self):
        self.getColorForButton(self.form.ButtonSectionColor)

    def getColorForButton(self, button):
        "opens a color picker dialog"
        icon = button.icon()
        pixel = icon.pixmap(16, 16).toImage().pixel(0, 0)
        color = QtGui.QColorDialog.getColor(QtGui.QColor(pixel))
        if color.isValid():
            colorPix = QtGui.QPixmap(16, 16)
            colorPix.fill(color)
            button.setIcon(QtGui.QIcon(colorPix))

    def fillMaterialCombo(self):
        "fills the combo with the existing FCMat cards"
        # look for cards in both resources dir and a Materials sub-folder in the user folder.
        # User cards with same name will override system cards
        resources_mat_path = os.path.join(
            FreeCAD.getResourceDir(), "Mod", "Material", "Resources", "Materials"
        )
        resources_mat_path_std = os.path.join(resources_mat_path, "Standard")
        user_mat_path = os.path.join(FreeCAD.ConfigGet("UserAppData"), "Material")

        paths = [resources_mat_path_std]
        if os.path.exists(user_mat_path):
            paths.append(user_mat_path)
        self.cards = {}
        for p in paths:
            for root, _, f_names in os.walk(p):
                for f in f_names:
                    b, e = os.path.splitext(f)
                    if e.upper() == ".FCMAT":
                        self.cards[b] = os.path.join(root, f)
        if self.cards:
            for k in sorted(self.cards):
                self.form.comboBox_MaterialsInDir.addItem(k)

    def fillExistingCombo(self):
        "fills the existing materials combo"
        self.existingmaterials = []
        for obj in FreeCAD.ActiveDocument.Objects:
            if obj.isDerivedFrom("App::MaterialObject"):
                if obj != self.obj:
                    self.existingmaterials.append(obj)
        for m in self.existingmaterials:
            self.form.comboBox_FromExisting.addItem(m.Label)

    def openEditor(self):
        "opens the full material editor from the material module"
        self.getFields()
        if self.material:
            import MaterialEditor

            self.material = MaterialEditor.editMaterial(self.material)
            self.setFields()

    def openUrl(self):
        self.getFields()
        if self.material:
            if "ProductURL" in self.material:
                QtGui.QDesktopServices.openUrl(self.material["ProductURL"])

    def getCode(self):
        FreeCADGui.Selection.addSelection(self.obj)
        FreeCADGui.runCommand("BIM_Classification")


class _ArchMultiMaterial:
    "The MultiMaterial object"

    def __init__(self, obj):
        self.Type = "MultiMaterial"
        obj.Proxy = self
        obj.addProperty(
            "App::PropertyString",
            "Description",
            "Arch",
            QT_TRANSLATE_NOOP("App::Property", "A description for this material"),
            locked=True,
        )
        obj.addProperty(
            "App::PropertyStringList",
            "Names",
            "Arch",
            QT_TRANSLATE_NOOP("App::Property", "The list of layer names"),
            locked=True,
        )
        obj.addProperty(
            "App::PropertyLinkList",
            "Materials",
            "Arch",
            QT_TRANSLATE_NOOP("App::Property", "The list of layer materials"),
            locked=True,
        )
        obj.addProperty(
            "App::PropertyFloatList",
            "Thicknesses",
            "Arch",
            QT_TRANSLATE_NOOP("App::Property", "The list of layer thicknesses"),
            locked=True,
        )

    def dumps(self):
        if hasattr(self, "Type"):
            return self.Type

    def loads(self, state):
        if state:
            self.Type = state


class _ViewProviderArchMultiMaterial:
    "A View Provider for the MultiMaterial object"

    def __init__(self, vobj):
        vobj.Proxy = self

    def getIcon(self):
        return ":/icons/Arch_Material_Multi.svg"

    def attach(self, vobj):
        self.Object = vobj.Object

    def setEdit(self, vobj, mode):
        if mode != 0:
            return None

        taskd = _ArchMultiMaterialTaskPanel(vobj.Object)
        FreeCADGui.Control.showDialog(taskd)
        return True

    def unsetEdit(self, vobj, mode):
        if mode != 0:
            return None

        FreeCADGui.Control.closeDialog()
        return True

    def doubleClicked(self, vobj):
        self.edit()
        return True

    def setupContextMenu(self, vobj, menu):
        if FreeCADGui.activeWorkbench().name() != "BIMWorkbench":
            return
        actionEdit = QtGui.QAction(translate("Arch", "Edit"), menu)
        actionEdit.triggered.connect(self.edit)
        menu.addAction(actionEdit)

    def edit(self):
        FreeCADGui.ActiveDocument.setEdit(self.Object, 0)

    def dumps(self):
        return None

    def loads(self, state):
        return None

    def isShow(self):
        return True


if FreeCAD.GuiUp:

    class MultiMaterialDelegate(QtGui.QStyledItemDelegate):

        def __init__(self, parent=None, *args):
            self.mats = []
            for obj in FreeCAD.ActiveDocument.Objects:
                if obj.isDerivedFrom("App::MaterialObject"):
                    self.mats.append(obj)
            QtGui.QStyledItemDelegate.__init__(self, parent, *args)

        def createEditor(self, parent, option, index):
            if index.column() == 0:
                editor = QtGui.QComboBox(parent)
                editor.setEditable(True)
            elif index.column() == 1:
                editor = QtGui.QComboBox(parent)
            elif index.column() == 2:
                ui = FreeCADGui.UiLoader()
                editor = ui.createWidget("Gui::InputField")
                editor.setSizePolicy(QtGui.QSizePolicy.Preferred, QtGui.QSizePolicy.Minimum)
                editor.setParent(parent)
            else:
                editor = QtGui.QLineEdit(parent)
            return editor

        def setEditorData(self, editor, index):
            if index.column() == 0:
                import ArchWindow

                editor.addItems([index.data()] + ArchWindow.WindowPartTypes)
            elif index.column() == 1:
                idx = -1
                for i, m in enumerate(self.mats):
                    editor.addItem(m.Label)
                    if m.Label == index.data():
                        idx = i
                editor.setCurrentIndex(idx)
            else:
                QtGui.QStyledItemDelegate.setEditorData(self, editor, index)

        def setModelData(self, editor, model, index):
            if index.column() == 0:
                if editor.currentIndex() == -1:
                    model.setData(index, "")
                else:
                    model.setData(index, editor.currentText())
            elif index.column() == 1:
                if editor.currentIndex() == -1:
                    model.setData(index, "")
                else:
                    model.setData(index, self.mats[editor.currentIndex()].Label)
            else:
                QtGui.QStyledItemDelegate.setModelData(self, editor, model, index)


class _ArchMultiMaterialTaskPanel:
    """The editmode TaskPanel for MultiMaterial objects"""

    def __init__(self, obj=None):
        self.obj = obj
        self.form = FreeCADGui.PySideUic.loadUi(":/ui/ArchMultiMaterial.ui")
        self.model = QtGui.QStandardItemModel()
        self.model.setHorizontalHeaderLabels(
            [
                translate("Arch", "Name"),
                translate("Arch", "Material"),
                translate("Arch", "Thickness"),
            ]
        )
        self.form.tree.setRootIsDecorated(False)  # remove 1st column's extra left margin
        self.form.tree.setModel(self.model)
        self.form.tree.setUniformRowHeights(True)
        self.form.tree.setItemDelegate(MultiMaterialDelegate())
        self.form.chooseCombo.currentIndexChanged.connect(self.fromExisting)
        self.form.addButton.pressed.connect(self.addLayer)
        self.form.upButton.pressed.connect(self.upLayer)
        self.form.downButton.pressed.connect(self.downLayer)
        self.form.delButton.pressed.connect(self.delLayer)
        self.form.invertButton.pressed.connect(self.invertLayer)
        self.model.itemChanged.connect(self.recalcThickness)
        self.fillExistingCombo()
        self.fillData()

    def fillData(self, obj=None):
        if not obj:
            obj = self.obj
        if obj:
            self.model.clear()
            self.model.setHorizontalHeaderLabels(
                [
                    translate("Arch", "Name"),
                    translate("Arch", "Material"),
                    translate("Arch", "Thickness"),
                ]
            )
            # restore widths
            self.form.tree.setColumnWidth(0, params.get_param_arch("MultiMaterialColumnWidth0"))
            self.form.tree.setColumnWidth(1, params.get_param_arch("MultiMaterialColumnWidth1"))
            for i in range(len(obj.Names)):
                item1 = QtGui.QStandardItem(obj.Names[i])
                item2 = QtGui.QStandardItem(obj.Materials[i].Label)
                item3 = QtGui.QStandardItem(
                    FreeCAD.Units.Quantity(
                        obj.Thicknesses[i], FreeCAD.Units.Length
                    ).getUserPreferred()[0]
                )
                self.model.appendRow([item1, item2, item3])
            self.form.nameField.setText(obj.Label)

    def fillExistingCombo(self):
        "fills the existing multimaterials combo"
        import Draft

        self.existingmaterials = []
        for obj in FreeCAD.ActiveDocument.Objects:
            if Draft.getType(obj) == "MultiMaterial":
                if obj != self.obj:
                    self.existingmaterials.append(obj)
        for m in self.existingmaterials:
            self.form.chooseCombo.addItem(m.Label)

    def fromExisting(self, index):
        "sets the contents from an existing material"
        if index > 0:
            if index <= len(self.existingmaterials):
                m = self.existingmaterials[index - 1]
                if m:
                    self.fillData(m)

    def addLayer(self):
        item1 = QtGui.QStandardItem(translate("Arch", "New layer"))
        item2 = QtGui.QStandardItem()
        item3 = QtGui.QStandardItem()
        self.model.appendRow([item1, item2, item3])

    def delLayer(self):
        sel = self.form.tree.selectedIndexes()
        if sel:
            row = sel[0].row()
            if row >= 0:
                self.model.takeRow(row)
        self.recalcThickness()

    def moveLayer(self, mvt=0):
        sel = self.form.tree.selectedIndexes()
        if sel and mvt:
            row = sel[0].row()
            if row >= 0:
                if row + mvt >= 0:
                    data = self.model.takeRow(row)
                    self.model.insertRow(row + mvt, data)
                    ind = self.model.index(row + mvt, 0)
                    self.form.tree.setCurrentIndex(ind)

    def upLayer(self):
        self.moveLayer(mvt=-1)

    def downLayer(self):
        self.moveLayer(mvt=1)

    def invertLayer(self):
        items = [self.model.takeRow(row) for row in range(self.model.rowCount() - 1, -1, -1)]
        items.reverse()
        for item in items:
            self.model.insertRow(0, item)

    def recalcThickness(self, item=None):
        prefix = translate("Arch", "Total thickness") + ": "
        th = 0
        suffix = ""
        for row in range(self.model.rowCount()):
            thick = 0
            d = self.model.item(row, 2).text()
            try:
                d = float(d)
            except Exception:
                thick = FreeCAD.Units.Quantity(d).Value
            else:
                thick = FreeCAD.Units.Quantity(d, FreeCAD.Units.Length).Value
            th += abs(thick)
            if not thick:
                suffix = " (" + translate("Arch", "depends on the object") + ")"
        val = FreeCAD.Units.Quantity(th, FreeCAD.Units.Length).UserString
        self.form.labelTotalThickness.setText(prefix + val + suffix)

    def accept(self):
        # store widths
        params.set_param_arch("MultiMaterialColumnWidth0", self.form.tree.columnWidth(0))
        params.set_param_arch("MultiMaterialColumnWidth1", self.form.tree.columnWidth(1))
        if self.obj:
            mats = []
            for m in FreeCAD.ActiveDocument.Objects:
                if m.isDerivedFrom("App::MaterialObject"):
                    mats.append(m)
            names = []
            materials = []
            thicknesses = []
            for row in range(self.model.rowCount()):
                name = self.model.item(row, 0).text()
                mat = None
                ml = self.model.item(row, 1).text()
                for m in mats:
                    if m.Label == ml:
                        mat = m
                d = self.model.item(row, 2).text()
                try:
                    d = float(d)
                except Exception:
                    thick = FreeCAD.Units.Quantity(d).Value
                else:
                    thick = FreeCAD.Units.Quantity(d, FreeCAD.Units.Length).Value
                if round(thick, 32) == 0:
                    thick = 0.0
                if name and mat:
                    names.append(name)
                    materials.append(mat)
                    thicknesses.append(thick)
            self.obj.Names = names
            self.obj.Materials = materials
            self.obj.Thicknesses = thicknesses
            if self.form.nameField.text():
                self.obj.Label = self.form.nameField.text()
        FreeCAD.ActiveDocument.recompute()
        FreeCADGui.ActiveDocument.resetEdit()
        return True

    def reject(self):
        FreeCADGui.ActiveDocument.resetEdit()
        return True



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2024 Mario Passaglia <mpassaglia[at]cbc.uba.ar>         *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM base element object"
__author__ = "Mario Passaglia"
__url__ = "https://www.freecad.org"

## @package base_femmeshelement
#  \ingroup FEM
#  \brief base object for FEM Mesh Element Features


from FreeCAD import Base
from . import base_fempythonobject

_PropHelper = base_fempythonobject._PropHelper


class BaseFemMeshElement(base_fempythonobject.BaseFemPythonObject):

    BaseType = "Fem::BaseFemMeshElement"

    def __init__(self, obj):
        super().__init__(obj)

        for prop in self._get_properties():
            prop.add_to_object(obj)

        obj.addExtension("App::SuppressibleExtensionPython")

    def _get_properties(self):
        prop = []

        prop.append(
            _PropHelper(
                type="App::PropertyLinkSubListGlobal",
                name="References",
                group="Mesh Element",
                doc="List of reference shapes",
                value=[],
            )
        )

        return prop

    def onDocumentRestored(self, obj):
        # update old project with new properties
        for prop in self._get_properties():
            try:
                obj.getPropertyByName(prop.name)
            except Base.PropertyError:
                prop.add_to_object(obj)

            if prop.name == "References":
                # change References to App::PropertyLinkSubListGlobal
                prop.handle_change_type(obj, old_type="App::PropertyLinkSubList")

        if not obj.hasExtension("App::SuppressibleExtensionPython"):
            obj.addExtension("App::SuppressibleExtensionPython")



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2019 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the task panel code for the Draft SelectPlane tool.

As it is right now this code only loads the task panel .ui file.
All logic on how to use the widgets is located in the GuiCommand class
itself.
On the other hand, the newer tools introduced in v0.19 like OrthoArray,
PolarArray, and CircularArray include the logic and manipulation
of the widgets in this task panel class.
In addition, the task panel code launches the actual function
using the delayed mechanism defined by the `todo.ToDo` class.
Therefore, at some point this class should be refactored
to be more similar to OrthoArray and the new tools.
"""
## @package task_selectplane
# \ingroup drafttaskpanels
# \brief Provides the task panel code for the Draft SelectPlane tool.

## \addtogroup drafttaskpanels
# @{
import FreeCADGui as Gui
from PySide import QtWidgets


class SelectPlaneTaskPanel:
    """The task panel definition of the Draft_SelectPlane command."""

    def __init__(self):
        self.form = Gui.PySideUic.loadUi(":/ui/TaskSelectPlane.ui")

    def getStandardButtons(self):
        """Execute to set the standard buttons."""
        return QtWidgets.QDialogButtonBox.Close


## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009 Yorik van Havre <yorik@uncreated.net>              *
# *   Copyright (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *   Copyright (c) 2025 FreeCAD Project Association                        *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the task panel code for the Draft Scale tool."""
## @package task_scale
# \ingroup drafttaskpanels
# \brief Provides the task panel code for the Draft Scale tool.

## \addtogroup drafttaskpanels
# @{
import PySide.QtCore as QtCore
import PySide.QtGui as QtGui
import PySide.QtWidgets as QtWidgets

import FreeCAD as App
import FreeCADGui as Gui
from draftutils import params
from draftutils import utils
from draftutils.translate import translate


class ScaleTaskPanel:
    """The task panel for the Draft Scale tool."""

    def __init__(self):
        decimals = max(6, params.get_param("Decimals", path="Units"))
        self.sourceCmd = None
        self.form = QtWidgets.QWidget()
        self.form.setWindowTitle(translate("Draft", "Scale"))
        self.form.setWindowIcon(QtGui.QIcon(":/icons/Draft_Scale.svg"))
        layout = QtWidgets.QGridLayout(self.form)
        self.xLabel = QtWidgets.QLabel()
        self.xLabel.setText(translate("Draft", "X-factor"))
        layout.addWidget(self.xLabel, 0, 0, 1, 1)
        self.xValue = QtWidgets.QDoubleSpinBox()
        self.xValue.setRange(-1000000.0, 1000000.0)
        self.xValue.setDecimals(decimals)
        self.xValue.setValue(1)
        layout.addWidget(self.xValue, 0, 1, 1, 1)
        self.yLabel = QtWidgets.QLabel()
        self.yLabel.setText(translate("Draft", "Y-factor"))
        layout.addWidget(self.yLabel, 1, 0, 1, 1)
        self.yValue = QtWidgets.QDoubleSpinBox()
        self.yValue.setRange(-1000000.0, 1000000.0)
        self.yValue.setDecimals(decimals)
        self.yValue.setValue(1)
        layout.addWidget(self.yValue, 1, 1, 1, 1)
        self.zLabel = QtWidgets.QLabel()
        self.zLabel.setText(translate("Draft", "Z-factor"))
        layout.addWidget(self.zLabel, 2, 0, 1, 1)
        self.zValue = QtWidgets.QDoubleSpinBox()
        self.zValue.setRange(-1000000.0, 1000000.0)
        self.zValue.setDecimals(decimals)
        self.zValue.setValue(1)
        layout.addWidget(self.zValue, 2, 1, 1, 1)
        self.lock = QtWidgets.QCheckBox()
        self.lock.setText(translate("Draft", "Uniform scaling"))
        self.lock.setChecked(params.get_param("ScaleUniform"))
        layout.addWidget(self.lock, 3, 0, 1, 2)

        QtCore.QObject.connect(self.xValue, QtCore.SIGNAL("valueChanged(double)"), self.setValue)
        QtCore.QObject.connect(self.yValue, QtCore.SIGNAL("valueChanged(double)"), self.setValue)
        QtCore.QObject.connect(self.zValue, QtCore.SIGNAL("valueChanged(double)"), self.setValue)
        QtCore.QObject.connect(self.lock, QtCore.SIGNAL("toggled(bool)"), self.setLock)

        if self.__class__.__name__ != "ScaleTaskPanelEdit":
            # ScaleRelative option removed in v1.1 as it does not work properly:
            # self.relative = QtWidgets.QCheckBox()
            # self.relative.setText(translate("Draft", "Working plane orientation"))
            # self.relative.setChecked(params.get_param("ScaleRelative"))
            # layout.addWidget(self.relative,4,0,1,2)
            self.isCopy = QtWidgets.QCheckBox()
            self.isCopy.setText(translate("Draft", "Copy"))
            self.isCopy.setChecked(params.get_param("ScaleCopy"))
            layout.addWidget(self.isCopy, 5, 0, 1, 2)
            self.isSubelementMode = QtWidgets.QCheckBox()
            self.isSubelementMode.setText(translate("Draft", "Modify subelements"))
            self.isSubelementMode.setChecked(params.get_param("SubelementMode"))
            layout.addWidget(self.isSubelementMode, 6, 0, 1, 2)
            self.isClone = QtWidgets.QCheckBox()
            self.isClone.setText(translate("Draft", "Create a clone"))
            self.isClone.setChecked(params.get_param("ScaleClone"))
            layout.addWidget(self.isClone, 7, 0, 1, 2)
            self.pickrefButton = QtWidgets.QPushButton()
            self.pickrefButton.setText(translate("Draft", "Pick From/To Points"))
            layout.addWidget(self.pickrefButton, 8, 0, 1, 2)

            # QtCore.QObject.connect(self.relative,QtCore.SIGNAL("toggled(bool)"),self.setRelative)
            QtCore.QObject.connect(self.isCopy, QtCore.SIGNAL("toggled(bool)"), self.setCopy)
            QtCore.QObject.connect(
                self.isSubelementMode, QtCore.SIGNAL("toggled(bool)"), self.setSubelementMode
            )
            QtCore.QObject.connect(self.isClone, QtCore.SIGNAL("toggled(bool)"), self.setClone)
            QtCore.QObject.connect(self.pickrefButton, QtCore.SIGNAL("clicked()"), self.pickRef)

    def setValue(self, val=None):
        """Set the value of the scale factors."""
        if self.lock.isChecked():
            if not self.xValue.hasFocus():
                self.xValue.setValue(val)
            if not self.yValue.hasFocus():
                self.yValue.setValue(val)
            if not self.zValue.hasFocus():
                self.zValue.setValue(val)
        if self.sourceCmd:
            # self.sourceCmd is always None for ScaleTaskPanelEdit
            self.sourceCmd.scale_ghosts(
                self.xValue.value(), self.yValue.value(), self.zValue.value()
            )

    def setLock(self, state):
        """Set the uniform scaling."""
        if self.sourceCmd:
            # self.sourceCmd is always None for ScaleTaskPanelEdit
            params.set_param("ScaleUniform", state)
        if state:
            val = self.xValue.value()
            self.yValue.setValue(val)
            self.zValue.setValue(val)

    def setRelative(self, state):
        """Set the relative scaling."""
        params.set_param("ScaleRelative", state)
        if self.sourceCmd:
            self.sourceCmd.scale_ghosts(
                self.xValue.value(), self.yValue.value(), self.zValue.value()
            )

    def setCopy(self, state):
        """Set the copy option."""
        params.set_param("ScaleCopy", state)
        if state and self.isClone.isChecked():
            self.isClone.setChecked(False)

    def setSubelementMode(self, state):
        """Set the subelement option."""
        params.set_param("SubelementMode", state)
        if state and self.isClone.isChecked():
            self.isClone.setChecked(False)
        if self.sourceCmd:
            self.sourceCmd.set_ghosts()
            self.sourceCmd.scale_ghosts(
                self.xValue.value(), self.yValue.value(), self.zValue.value()
            )

    def setClone(self, state):
        """Set the clone option."""
        params.set_param("ScaleClone", state)
        if state and self.isCopy.isChecked():
            self.isCopy.setChecked(False)
        if state and self.isSubelementMode.isChecked():
            self.isSubelementMode.setChecked(False)

    def pickRef(self):
        """Pick a reference point from the calling class."""
        if self.sourceCmd:
            self.sourceCmd.pick_ref()

    def accept(self):
        """Execute when clicking the OK button."""
        if self.sourceCmd:
            self.sourceCmd.scale()
        Gui.ActiveDocument.resetEdit()
        return True

    def reject(self):
        """Execute when clicking the Cancel button."""
        if self.sourceCmd:
            self.sourceCmd.finish()
        Gui.ActiveDocument.resetEdit()
        return True


class ScaleTaskPanelEdit(ScaleTaskPanel):
    """The task panel to edit the scale of Draft Clones."""

    def __init__(self, obj):
        super().__init__()
        self.ghost = None
        self.selection = Gui.Selection.getSelectionEx("", 0)
        self.obj = obj
        self.obj_x, self.obj_y, self.obj_z = self.obj.Scale
        self.form.setWindowTitle(translate("Draft", "Edit Scale"))
        self.form.setWindowIcon(QtGui.QIcon(":/icons/Draft_Clone.svg"))
        self.xValue.setValue(self.obj_x)
        self.yValue.setValue(self.obj_y)
        self.zValue.setValue(self.obj_z)
        self.lock.setChecked(self.obj_x == self.obj_y == self.obj_z)

    def setValue(self, val=None):
        """Set the value of the scale factors."""
        super().setValue(val)
        self.scale_ghost(self.xValue.value(), self.yValue.value(), self.zValue.value())

    def scale_ghost(self, x, y, z):
        """Scale the preview of the object."""
        x = x / (self.obj_x if abs(self.obj_x) > 1e-7 else 1e-7)
        y = y / (self.obj_y if abs(self.obj_y) > 1e-7 else 1e-7)
        z = z / (self.obj_z if abs(self.obj_z) > 1e-7 else 1e-7)

        if self.ghost is None:
            self.set_ghost()

        mtx_scale = App.Matrix()
        mtx_scale.scale(x, y, z)
        mtx = self.global_place.Matrix * mtx_scale
        mtx = mtx * self.global_place.Matrix.inverse()

        delta = self.global_place.inverse().Rotation.multVec(self.global_place.Base)
        delta = -App.Vector(delta.x * x, delta.y * y, delta.z * z)
        delta = self.global_place.multVec(delta)

        self.ghost.setMatrix(mtx)
        self.ghost.move(delta)
        # self.ghost.flip_normals(x * y * z < 0)  # Does not work properly for Draft_Circles for example.
        self.ghost.on()

    def set_ghost(self):
        """Set the ghost to display."""
        # Import has to happen here to avoid circular imports.
        from draftguitools import gui_trackers as trackers

        if self.ghost is not None:
            self.ghost.remove()
        objs, places, _ = utils._modifiers_process_selection(self.selection, copy=False, scale=True)
        self.ghost = trackers.ghostTracker(objs, parent_places=places)
        self.global_place = places[0] * self.obj.Placement

    def accept(self):
        """Execute when clicking the OK button."""
        self.obj.Scale = (self.xValue.value(), self.yValue.value(), self.zValue.value())
        App.ActiveDocument.recompute()
        if self.ghost is not None:
            self.ghost.finalize()
        Gui.ActiveDocument.resetEdit()
        return True

    def reject(self):
        """Execute when clicking the Cancel button."""
        if self.ghost is not None:
            self.ghost.finalize()
        Gui.ActiveDocument.resetEdit()
        return True

    def finish(self):
        """Called by unsetEdit in view_clone.py."""
        Gui.Control.closeDialog()
        return None


## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2002 Juergen Riegel <juergen.riegel@web.de>             *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Lesser General Public License for more details.                   *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************/

# ReverseEngineering gui init module
#
# Gathering all the information to start FreeCAD
# This is the second one of three init scripts, the third one
# runs when the gui is up


class ReverseEngineeringWorkbench(Workbench):
    "ReverseEngineering workbench object"

    def __init__(self):
        self.__class__.Icon = (
            FreeCAD.getResourceDir()
            + "Mod/ReverseEngineering/Resources/icons/ReverseEngineeringWorkbench.svg"
        )
        self.__class__.MenuText = "Reverse Engineering"
        self.__class__.ToolTip = "Reverse Engineering workbench"

    def Initialize(self):
        # load the module
        import ReverseEngineeringGui
        import ReverseEngineering

    def GetClassName(self):
        return "ReverseEngineeringGui::Workbench"


Gui.addWorkbench(ReverseEngineeringWorkbench())



#!/usr/bin/env python3
# SPDX-License-Identifier: EUPL-1.2

# ***************************************************************************
# *  Copyright (c) 2025 Clair-Loup Sergent <clsergent@free.fr>              *
# *                                                                         *
# *  Licensed under the EUPL-1.2 with the specific provision                *
# *  (EUPL articles 14 & 15) that the applicable law is the French law.     *
# *  and the Jurisdiction Paris.                                            *
# *  Any redistribution must include the specific provision above.          *
# *                                                                         *
# *  You may obtain a copy of the Licence at:                               *
# *  https://joinup.ec.europa.eu/software/page/eupl5                        *
# *                                                                         *
# *  Unless required by applicable law or agreed to in writing, software    *
# *  distributed under the Licence is distributed on an "AS IS" basis,      *
# *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or        *
# *  implied. See the Licence for the specific language governing           *
# *  permissions and limitations under the Licence.                         *
# ***************************************************************************
import argparse
import re
from typing import List

import FreeCAD

import Path
import CAMTests.PathTestUtils as PathTestUtils
from Path.Post.Processor import PostProcessorFactory

Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
Path.Log.trackModule(Path.Log.thisModule())


class TestSnapmakerPost(PathTestUtils.PathTestBase):
    """Test the Snapmaker postprocessor."""

    @classmethod
    def setUpClass(cls):
        """Set up the test environment"""

        FreeCAD.ConfigSet("SuppressRecomputeRequiredDialog", "True")
        cls.doc = FreeCAD.open(FreeCAD.getHomePath() + "/Mod/CAM/CAMTests/boxtest.fcstd")
        cls.job = cls.doc.getObject("Job")
        cls.post = PostProcessorFactory.get_post_processor(cls.job, "snapmaker")
        # locate the operation named "Profile"
        for op in cls.job.Operations.Group:
            if op.Label == "Profile":
                # remember the "Profile" operation
                cls.profile_op = op
                return

    @classmethod
    def tearDownClass(cls):
        """Tear down the test environment"""
        FreeCAD.closeDocument(cls.doc.Name)
        FreeCAD.ConfigSet("SuppressRecomputeRequiredDialog", "")

    def setUp(self):
        """Unit test init"""
        # allow a full length "diff" if an error occurs
        self.maxDiff = None
        # reinitialize the postprocessor data structures between tests
        self.post.initialize()

    def tearDown(self):
        """Unit test tear down"""
        pass

    def get_gcode(self, ops: List[str], arguments: str) -> str:
        """Get postprocessed gcode from a list of operations and postprocessor arguments"""
        self.profile_op.Path = Path.Path(ops)
        self.job.PostProcessorArgs = "--no-show-editor --no-gui --no-thumbnail " + arguments
        return self.post.export()[0][1]

    def test_general(self):
        """Test Output Generation"""

        expected_header = """\
;Header Start
;header_type: cnc
;machine: Snapmaker 2 A350 50W CNC module
;Post Processor: snapmaker_post
;CAM File: boxtest.fcstd
;Output Time: \\d{4}-\\d{2}-\\d{2} \\d{2}:\\d{2}:\\d{2}\\.\\d{0,6}
;thumbnail: deactivated."""

        expected_body = """\
;Begin preamble
G90
G17
G21
;Begin operation: Fixture
;Path: Fixture
G54
;End operation: Fixture
;Begin operation: TC: Default Tool
;Path: TC: Default Tool
;TC: Default Tool
;Begin toolchange
M5
M76
M6 T1
;End operation: TC: Default Tool
;Begin operation: Profile
;Path: Profile
;End operation: Profile
;Begin postamble
M400
M5
"""

        # test header and body with comments
        gcode = self.get_gcode([], "--machine=A350 --toolhead=50W_CNC")

        g_lines = gcode.splitlines()
        e_lines = expected_header.splitlines() + expected_body.splitlines()

        self.assertTrue(len(g_lines), len(e_lines))
        for (nbr, exp), line in zip(enumerate(e_lines), g_lines):
            if exp.startswith(";Output Time:"):
                self.assertTrue(re.match(exp, line) is not None)
            else:
                self.assertEqual(exp, line)

        # test body without header
        gcode = self.get_gcode([], "--machine=A350 --toolhead=50W_CNC --no-header")
        self.assertEqual(gcode, expected_body)

        # test body without comments
        gcode = self.get_gcode([], "--machine=A350 --toolhead=50W_CNC --no-header --no-comments")
        expected = "".join(
            [line for line in expected_body.splitlines(keepends=True) if not line.startswith(";")]
        )
        self.assertEqual(gcode, expected)

    def test_command(self):
        """Test command Generation"""
        command = Path.Command("G0 X10 Y20 Z30")
        expected = "G0 X10.000 Y20.000 Z30.000"

        gcode = self.get_gcode([command], "--machine=A350 --toolhead=50W_CNC --no-header")
        result = gcode.splitlines()[18]
        self.assertEqual(result, expected)

    def test_precision(self):
        """Test Precision"""
        # test G0 command with precision 2 digits precision
        command = Path.Command("G0 X10 Y20 Z30")
        expected = "G0 X10.00 Y20.00 Z30.00"

        gcode = self.get_gcode(
            [command], "--machine=A350 --toolhead=50W_CNC --no-header --precision=2"
        )
        result = gcode.splitlines()[18]
        self.assertEqual(result, expected)

    def test_lines(self):
        """Test Line Numbers"""
        command = Path.Command("G0 X10 Y20 Z30")
        expected = "N46 G0 X10.000 Y20.000 Z30.000"

        gcode = self.get_gcode(
            [command],
            "--machine=A350 --toolhead=50W_CNC --no-header --line-numbers --line-number=10 --line-increment=2",
        )
        result = gcode.splitlines()[18]
        self.assertEqual(result, expected)

    def test_preamble(self):
        """Test Pre-amble"""
        gcode = self.get_gcode(
            [],
            "--machine=A350 --toolhead=50W_CNC --no-header --preamble='G18 G55' --no-comments",
        )
        result = gcode.splitlines()[0]
        self.assertEqual(result, "G18 G55")

    def test_postamble(self):
        """Test Post-amble"""
        gcode = self.get_gcode(
            [],
            "--machine=A350 --toolhead=50W_CNC --no-header --postamble='G0 Z50\nM2' --no-comments",
        )
        result = gcode.splitlines()[-2]
        self.assertEqual(result, "G0 Z50")
        self.assertEqual(gcode.splitlines()[-1], "M2")

    def test_inches(self):
        """Test inches conversion"""

        command = Path.Command("G0 X10 Y20 Z30")

        # test inches conversion
        expected = "G0 X0.3937 Y0.7874 Z1.1811"
        gcode = self.get_gcode([command], "--machine=A350 --toolhead=50W_CNC --no-header --inches")
        self.assertEqual(gcode.splitlines()[3], "G20")
        result = gcode.splitlines()[18]
        self.assertEqual(result, expected)

        # test inches conversion with 2 digits precision
        expected = "G0 X0.39 Y0.79 Z1.18"
        gcode = self.get_gcode(
            [command],
            "--machine=A350 --toolhead=50W_CNC --no-header --inches --precision=2",
        )
        result = gcode.splitlines()[18]
        self.assertEqual(result, expected)

    def test_axis_modal(self):
        """Test axis modal - Suppress the axis coordinate if the same as previous"""

        c0 = Path.Command("G0 X10 Y20 Z30")
        c1 = Path.Command("G0 X10 Y30 Z30")
        expected = "G0 Y30.000"

        gcode = self.get_gcode(
            [c0, c1], "--machine=A350 --toolhead=50W_CNC --no-header --axis-modal"
        )
        result = gcode.splitlines()[19]
        self.assertEqual(result, expected)

    def test_tool_change(self):
        """Test tool change"""

        c0 = Path.Command("M6 T2")
        c1 = Path.Command("M3 S3000")

        gcode = self.get_gcode([c0, c1], "--machine=A350 --toolhead=50W_CNC --no-header")
        self.assertEqual(gcode.splitlines()[19:22], ["M5", "M76", "M6 T2"])
        self.assertEqual(
            gcode.splitlines()[22], "M3 P25"
        )  # no TLO on Snapmaker (G43 inserted after tool change)

    def test_models(self):
        """Test the various models, and also test models that don't exist cause an error."""
        command = Path.Command("G0 X10 Y20 Z30")
        expected = "G0 X10.000 Y20.000 Z30.000"

        with self.assertRaises(SystemExit):
            self.get_gcode(
                [command],
                "--no-header",
            )

        with self.assertRaises(SystemExit):
            gcode = self.get_gcode(
                [command],
                "--machine=robot --no-header",
            )

        gcode = self.get_gcode(
            [command],
            "--machine=Original --no-header",
        )
        result = gcode.splitlines()[18]
        self.assertEqual(result, expected)

        gcode = self.get_gcode(
            [command],
            "--machine=A150 --toolhead=50W_CNC --no-header",
        )
        result = gcode.splitlines()[18]
        self.assertEqual(result, expected)

        gcode = self.get_gcode(
            [command],
            "--machine=A250 --toolhead=50W_CNC --no-header",
        )
        result = gcode.splitlines()[18]
        self.assertEqual(result, expected)

        gcode = self.get_gcode(
            [command],
            "--machine=A250T --toolhead=50W_CNC --no-header",
        )
        result = gcode.splitlines()[18]
        self.assertEqual(result, expected)

        gcode = self.get_gcode(
            [command],
            "--machine=A250T --toolhead=200W_CNC --bracing-kit --no-header",
        )
        result = gcode.splitlines()[18]
        self.assertEqual(result, expected)

        gcode = self.get_gcode(
            [command],
            "--machine=A350 --toolhead=50W_CNC --no-header",
        )
        result = gcode.splitlines()[18]
        self.assertEqual(result, expected)

        gcode = self.get_gcode(
            [command],
            "--machine=A350T --toolhead=50W_CNC --no-header",
        )
        result = gcode.splitlines()[18]
        self.assertEqual(result, expected)

        gcode = self.get_gcode(
            [command],
            "--machine=A350T --toolhead=200W_CNC --bracing-kit --no-header",
        )
        result = gcode.splitlines()[18]
        self.assertEqual(result, expected)

        gcode = self.get_gcode(
            [command],
            "--machine=Artisan --no-header",
        )
        result = gcode.splitlines()[18]
        self.assertEqual(result, expected)

    def test_mod_kits(self):
        """Test the various mod kits against various models."""

        # Reference for boundaries with the bracing kit and quick swap kit combinations
        # [1] https://support.snapmaker.com/hc/en-us/articles/20786910972311-FAQ-for-Bracing-Kit-for-Snapmaker-2-0-Linear-Modules#h_01HN4Z7S9WJE5BRT492WR0CKH1
        # Reference for quick swap kit
        # [2] https://support.snapmaker.com/hc/en-us/articles/15320624494103-Pre-sale-FAQ-for-Quick-Swap-Kit

        command = Path.Command("G0 X10 Y20 Z30")
        expected = "G0 X10.000 Y20.000 Z30.000"

        gcode = self.get_gcode(
            [command],
            "--machine=Original --no-header",
        )
        result = gcode.splitlines()[18]
        self.assertEqual(result, expected)
        self.assertEqual(self.post.values["MOD_KITS_INSTALLED"], [])
        # https://forum.snapmaker.com/t/cnc-work-area-size/5178
        self.assertEqual(self.post.values["BOUNDARIES"], dict(X=125, Y=125, Z=50))

        gcode = self.get_gcode(
            [command],
            "--machine=Original --quick-swap --no-header",
        )
        # I don't understand why export returns the arguments
        # if snapmaker_process_arguments fails.
        self.assertTrue(isinstance(gcode, argparse.Namespace))
        self.assertFalse(isinstance(gcode, str))

        gcode = self.get_gcode(
            [command],
            "--machine=Original --bracing-kit --no-header",
        )
        # I don't understand why export returns the arguments
        # if snapmaker_process_arguments fails.
        self.assertTrue(isinstance(gcode, argparse.Namespace))
        self.assertFalse(isinstance(gcode, str))

        # This is incompatible according to [2]
        gcode = self.get_gcode(
            [command],
            "--machine=A150 --quick-swap --no-header",
        )
        # I don't understand why export returns the arguments
        # if snapmaker_process_arguments fails.
        self.assertTrue(isinstance(gcode, argparse.Namespace))
        self.assertFalse(isinstance(gcode, str))

        gcode = self.get_gcode(
            [command],
            "--machine=Artisan --no-header",
        )
        result = gcode.splitlines()[18]
        self.assertEqual(result, expected)
        self.assertEqual(self.post.values["MOD_KITS_INSTALLED"], [])
        self.assertEqual(self.post.values["BOUNDARIES"], dict(X=400, Y=400, Z=400))

        gcode = self.get_gcode(
            [command],
            "--machine=Artisan --quick-swap --no-header",
        )
        # I don't understand why export returns the arguments
        # if snapmaker_process_arguments fails.
        self.assertTrue(isinstance(gcode, argparse.Namespace))
        self.assertFalse(isinstance(gcode, str))

        gcode = self.get_gcode(
            [command],
            "--machine=Artisan --bracing-kit --no-header",
        )
        # I don't understand why export returns the arguments
        # if snapmaker_process_arguments fails.
        self.assertTrue(isinstance(gcode, argparse.Namespace))
        self.assertFalse(isinstance(gcode, str))

        # This test case is covered in reference [1]
        gcode = self.get_gcode(
            [command],
            "--machine=A150 --toolhead=50W_CNC --no-header",
        )
        result = gcode.splitlines()[18]
        self.assertEqual(result, expected)
        self.assertEqual(self.post.values["MOD_KITS_INSTALLED"], [])
        self.assertEqual(self.post.values["BOUNDARIES"], dict(X=145, Y=160, Z=90))

        # This test case is covered in reference [1]
        gcode = self.get_gcode(
            [command],
            "--machine=A150 --toolhead=50W_CNC --bracing-kit --no-header",
        )
        result = gcode.splitlines()[18]
        self.assertEqual(result, expected)
        self.assertEqual(self.post.values["MOD_KITS_INSTALLED"], ["BK"])
        self.assertEqual(self.post.values["BOUNDARIES"], dict(X=145, Y=148, Z=90))

        # This test case is covered in reference [1]
        gcode = self.get_gcode(
            [command],
            "--machine=A250 --toolhead=50W_CNC --bracing-kit --quick-swap --no-header",
        )
        result = gcode.splitlines()[18]
        self.assertEqual(result, expected)
        self.assertEqual(self.post.values["MOD_KITS_INSTALLED"], ["QS", "BK"])
        self.assertEqual(self.post.values["BOUNDARIES"], dict(X=230, Y=223, Z=180))

        # This test case is covered in reference [1]
        gcode = self.get_gcode(
            [command],
            "--machine=A250T --toolhead=50W_CNC --quick-swap --no-header",
        )
        result = gcode.splitlines()[18]
        self.assertEqual(result, expected)
        self.assertEqual(self.post.values["MOD_KITS_INSTALLED"], ["QS"])
        self.assertEqual(self.post.values["BOUNDARIES"], dict(X=230, Y=235, Z=180))

        # This test case is covered in reference [1]
        gcode = self.get_gcode(
            [command],
            "--machine=A250T --toolhead=200W_CNC --bracing-kit --no-header",
        )
        result = gcode.splitlines()[18]
        self.assertEqual(result, expected)
        self.assertEqual(self.post.values["MOD_KITS_INSTALLED"], ["BK"])
        self.assertEqual(self.post.values["BOUNDARIES"], dict(X=230, Y=225, Z=180))

        # This test case is covered in reference [1]
        gcode = self.get_gcode(
            [command],
            "--machine=A350 --toolhead=50W_CNC --bracing-kit --no-header",
        )
        result = gcode.splitlines()[18]
        self.assertEqual(result, expected)
        self.assertEqual(self.post.values["MOD_KITS_INSTALLED"], ["BK"])
        self.assertEqual(self.post.values["BOUNDARIES"], dict(X=320, Y=338, Z=275))

        # This test case is covered in reference [1]
        gcode = self.get_gcode(
            [command],
            "--machine=A350 --toolhead=50W_CNC --quick-swap --no-header",
        )
        result = gcode.splitlines()[18]
        self.assertEqual(result, expected)
        self.assertEqual(self.post.values["MOD_KITS_INSTALLED"], ["QS"])
        self.assertEqual(self.post.values["BOUNDARIES"], dict(X=320, Y=335, Z=275))

        # This test case is covered in reference [1]
        gcode = self.get_gcode(
            [command],
            "--machine=A350 --toolhead=50W_CNC --bracing-kit --quick-swap --no-header",
        )
        result = gcode.splitlines()[18]
        self.assertEqual(result, expected)
        self.assertEqual(self.post.values["MOD_KITS_INSTALLED"], ["QS", "BK"])
        self.assertEqual(self.post.values["BOUNDARIES"], dict(X=320, Y=323, Z=275))

        # This test case is covered in reference [1]
        gcode = self.get_gcode(
            [command],
            "--machine=A350T --toolhead=50W_CNC --bracing-kit --quick-swap --no-header",
        )
        result = gcode.splitlines()[18]
        self.assertEqual(result, expected)
        self.assertEqual(self.post.values["MOD_KITS_INSTALLED"], ["QS", "BK"])
        self.assertEqual(self.post.values["BOUNDARIES"], dict(X=320, Y=323, Z=275))

        # This test case is covered in reference [1]
        gcode = self.get_gcode(
            [command],
            "--machine=A350T --toolhead=200W_CNC --bracing-kit --no-header",
        )
        result = gcode.splitlines()[18]
        self.assertEqual(result, expected)
        self.assertEqual(self.post.values["MOD_KITS_INSTALLED"], ["BK"])
        self.assertEqual(self.post.values["BOUNDARIES"], dict(X=320, Y=325, Z=275))

        # This test case is covered in reference [1]
        gcode = self.get_gcode(
            [command],
            "--machine=A350T --toolhead=200W_CNC --bracing-kit --quick-swap --no-header",
        )
        result = gcode.splitlines()[18]
        self.assertEqual(result, expected)
        self.assertEqual(self.post.values["MOD_KITS_INSTALLED"], ["QS", "BK"])
        self.assertEqual(self.post.values["BOUNDARIES"], dict(X=320, Y=310, Z=275))

    def test_toolhead_selection(self):
        """Test automatic selection of toolhead where appropriate"""

        # check succeeds
        command = Path.Command("G0 X10 Y20 Z30")
        expected = "G0 X10.000 Y20.000 Z30.000"

        gcode = self.get_gcode(
            [command],
            "--machine=Original --no-header",
        )
        result = gcode.splitlines()[18]
        self.assertEqual(result, expected)
        self.assertEqual(self.post.values["TOOLHEAD_NAME"], "Original CNC module")

        gcode = self.get_gcode(
            [command],
            "--machine=A350 --no-header",
        )
        self.assertFalse(isinstance(gcode, str))

        gcode = self.get_gcode(
            [command],
            "--machine=A350T --no-header",
        )
        self.assertFalse(isinstance(gcode, str))

        # check succeed with artisan (which base is bigger)
        gcode = self.get_gcode(
            [command],
            "--machine=Artisan --no-header --boundaries-check",
        )
        result = gcode.splitlines()[18]
        self.assertEqual(result, expected)
        self.assertEqual(self.post.values["TOOLHEAD_NAME"], "200W CNC module")

    def test_spindle_percent_rpm_auto_select(self):
        """Test automatic selection of spindle speed rpm vs percent"""

        command = Path.Command("M3 S2100")

        # test original toolhead
        gcode = self.get_gcode([command], "--machine=Original --no-header")
        self.assertEqual(gcode.splitlines()[18], "M3 P30")

        command = Path.Command("M3 S3600")

        # test 50W toolhead
        gcode = self.get_gcode([command], "--machine=A350 --toolhead=50W_CNC --no-header")
        self.assertEqual(gcode.splitlines()[18], "M3 P30")

        # test 200W toolhead
        gcode = self.get_gcode(
            [command], "--machine=A350 --toolhead=200W_CNC --bracing-kit --no-header"
        )
        self.assertEqual(gcode.splitlines()[18], "M3 S3600")

        # test 200W toolhead
        gcode = self.get_gcode([command], "--machine=Artisan --no-header")
        self.assertEqual(gcode.splitlines()[18], "M3 S3600")

    def test_spindle_percent(self):
        """Test spindle speed conversion from RPM to percents"""

        command = Path.Command("M3 S3600")

        # test 50W toolhead
        gcode = self.get_gcode(
            [command], "--machine=A350 --toolhead=50W_CNC --spindle-percent --no-header"
        )
        self.assertEqual(gcode.splitlines()[18], "M3 P30")

        # test 200W toolhead
        gcode = self.get_gcode(
            [command],
            "--machine=A350 --toolhead=200W_CNC --bracing-kit --spindle-percent --no-header",
        )
        self.assertEqual(gcode.splitlines()[18], "M3 P20")

        # test custom spindle speed extrema
        gcode = self.get_gcode(
            [command],
            "--machine=A350 --toolhead=200W_CNC --bracing-kit --spindle-percent --no-header --spindle-speeds=3000,4000",
        )
        self.assertEqual(gcode.splitlines()[18], "M3 P90")

    def test_comment(self):
        """Test comment"""

        command = Path.Command("(comment)")
        gcode = self.get_gcode(
            [command], "--machine=A350 --toolhead=50W_CNC --spindle-percent --no-header"
        )
        result = gcode.splitlines()[18]
        expected = ";comment"
        self.assertEqual(result, expected)

    def test_boundaries(self):
        """Test boundaries check"""

        # check succeeds
        command = Path.Command("G0 X100 Y-100.5 Z-1")

        gcode = self.get_gcode(
            [command],
            "--machine=A350 --toolhead=50W_CNC --no-header --boundaries-check",
        )
        self.assertTrue(self.post.check_boundaries(gcode.splitlines()))

        # check fails with A350
        c0 = Path.Command("G01 X100 Y-100.5 Z-1")
        c1 = Path.Command("G02 Y260")
        gcode = self.get_gcode(
            [c0, c1],
            "--machine=A350 --toolhead=50W_CNC --no-header --boundaries-check",
        )
        self.assertFalse(self.post.check_boundaries(gcode.splitlines()))

        # check succeed with artisan (which base is bigger)
        gcode = self.get_gcode(
            [c0, c1],
            "--machine=Artisan --no-header --boundaries-check",
        )
        self.assertTrue(self.post.check_boundaries(gcode.splitlines()))

        # check fails with custom boundaries
        gcode = self.get_gcode(
            [c0, c1],
            "--machine=A350 --toolhead=50W_CNC --no-header --boundaries-check --boundaries='50,400,10'",
        )
        self.assertFalse(self.post.check_boundaries(gcode.splitlines()))



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2011, 2012 Jose Luis Cercos Pita <jlcercos@gmail.com>   *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD

import PySide
from PySide import QtCore, QtGui
import sys

try:
    import matplotlib

    matplotlib.use("QtAgg")

    # Force matplotlib to use PySide backend by temporarily unloading PyQt
    pyqt5_unloaded = False
    pyqt6_unloaded = False
    if "PyQt5.QtCore" in sys.modules:
        del sys.modules["PyQt5.QtCore"]
        pyqt5_unloaded = True
    if "PyQt6.QtCore" in sys.modules:
        del sys.modules["PyQt6.QtCore"]
        pyqt6_unloaded = True

    import matplotlib.pyplot as plt

    # restore PyQt if unloaded
    if pyqt5_unloaded:
        import PyQt5.QtCore
    if pyqt6_unloaded:
        import PyQt6.QtCore

    from matplotlib.backends.backend_qtagg import FigureCanvasQTAgg as FigureCanvas
    from matplotlib.backends.backend_qtagg import NavigationToolbar2QT as NavigationToolbar

    from matplotlib.figure import Figure
except ImportError:
    FreeCAD.Console.PrintWarning(
        "The 'matplotlib' Python package was not found. Plot module cannot be loaded\n"
    )
    raise ImportError("matplotlib not installed")


def getMainWindow():
    """Return the FreeCAD main window."""
    toplevel = PySide.QtGui.QApplication.topLevelWidgets()
    for i in toplevel:
        if i.metaObject().className() == "Gui::MainWindow":
            return i
    return None


def getMdiArea():
    """Return FreeCAD MdiArea."""
    mw = getMainWindow()
    if not mw:
        return None
    childs = mw.children()
    for c in childs:
        if isinstance(c, PySide.QtGui.QMdiArea):
            return c
    return None


def getPlot():
    """Return the selected Plot document if exist."""
    # Get active tab
    mdi = getMdiArea()
    if not mdi:
        return None
    sub = mdi.activeSubWindow()
    if not sub:
        return None
    # Explore childrens looking for Plot class
    for i in sub.children():
        if i.metaObject().className() == "Plot":
            return i
    return None


def closePlot():
    """closePlot(): Close the active plot window."""
    # Get active tab
    mdi = getMdiArea()
    if not mdi:
        return None
    sub = mdi.activeSubWindow()
    if not sub:
        return None
    # Explore childrens looking for Plot class
    for i in sub.children():
        if i.metaObject().className() == "Plot":
            sub.close()


def figure(winTitle="plot"):
    """Create a new plot subwindow/tab.

    Keyword arguments:
    winTitle -- Plot tab title.
    """
    mdi = getMdiArea()
    if not mdi:
        return None
    win = Plot(winTitle)
    sub = mdi.addSubWindow(win)
    sub.show()
    return win


def plot(x, y, name=None):
    """Plots a new serie (as line plot)

    Keyword arguments:
    x -- X values
    y -- Y values
    name -- Data serie name (for legend).
    """
    # Get active plot, or create another one if don't exist
    plt = getPlot()
    if not plt:
        plt = figure()
    # Call to plot
    return plt.plot(x, y, name)


def series():
    """Return all the lines from a selected plot."""
    plt = getPlot()
    if not plt:
        return []
    return plt.series


def removeSerie(index):
    """Remove a data serie from the active plot.

    Keyword arguments:
    index -- Index of the serie to remove.
    """
    # Get active series
    plt = getPlot()
    if not plt:
        return
    plots = plt.series
    if not plots:
        return
    # Remove line from plot
    axes = plots[index].axes
    axes.lines.pop(plots[index].lid)
    # Remove serie from list
    del plt.series[index]
    # Update GUI
    plt.update()


def legend(status=True, pos=None, fontsize=None):
    """Show/Hide the legend from the active plot.

    Keyword arguments:
    status -- True if legend must be shown, False otherwise.
    pos -- Legend position.
    fontsize -- Font size
    """
    plt = getPlot()
    if not plt:
        return
    plt.legend = status
    if fontsize:
        plt.legSiz = fontsize
    # Hide all legends
    for axes in plt.axesList:
        axes.legend_ = None
    # Legend must be activated on last axes
    axes = plt.axesList[-1]
    if status:
        # Setup legend handles and names
        lines = series()
        handles = []
        names = []
        for l in lines:
            if l.name is not None:
                handles.append(l.line)
                names.append(l.name)
        # Show the legend (at selected position or at best)
        if pos:
            l = axes.legend(handles, names, bbox_to_anchor=pos)
            plt.legPos = pos
        else:
            l = axes.legend(handles, names, loc="best")
            # Update canvas in order to compute legend data
            plt.canvas.draw()
            # Get resultant position
            try:
                fax = axes.get_frame().get_extents()
            except Exception:
                fax = axes.patch.get_extents()
            fl = l.get_frame()
            plt.legPos = (
                (fl._x + fl._width - fax.x0) / fax.width,
                (fl._y + fl._height - fax.y0) / fax.height,
            )
        # Set fontsize
        for t in l.get_texts():
            t.set_fontsize(plt.legSiz)
    plt.update()


def grid(status=True):
    """Show/Hide the grid from the active plot.

    Keyword arguments:
    status -- True if grid must be shown, False otherwise.
    """
    plt = getPlot()
    if not plt:
        return
    plt.grid = status
    axes = plt.axes
    axes.grid(status)
    plt.update()


def title(string):
    """Setup the plot title.

    Keyword arguments:
    string -- Plot title.
    """
    plt = getPlot()
    if not plt:
        return
    axes = plt.axes
    axes.set_title(string)
    plt.update()


def xlabel(string):
    """Setup the x label.

    Keyword arguments:
    string -- Title to set.
    """
    plt = getPlot()
    if not plt:
        return
    axes = plt.axes
    axes.set_xlabel(string)
    plt.update()


def ylabel(string):
    """Setup the y label.

    Keyword arguments:
    string -- Title to set.
    """
    plt = getPlot()
    if not plt:
        return
    axes = plt.axes
    axes.set_ylabel(string)
    plt.update()


def axesList():
    """Return the plot axes sets list."""
    plt = getPlot()
    if not plt:
        return []
    return plt.axesList


def axes():
    """Return the active plot axes."""
    plt = getPlot()
    if not plt:
        return None
    return plt.axes


def addNewAxes(rect=None, frameon=True, patchcolor="none"):
    """Add new axes to plot, setting it as the active one.

    Keyword arguments:
    rect -- Axes area, None to copy from the last axes data.
    frameon -- True to show frame, False otherwise.
    patchcolor -- Patch color, 'none' for transparent plot.
    """
    plt = getPlot()
    if not plt:
        return None
    fig = plt.fig
    if rect is None:
        rect = plt.axes.get_position()
    ax = fig.add_axes(rect, frameon=frameon)
    ax.xaxis.set_ticks_position("bottom")
    ax.spines["top"].set_color("none")
    ax.yaxis.set_ticks_position("left")
    ax.spines["right"].set_color("none")
    ax.patch.set_facecolor(patchcolor)
    plt.axesList.append(ax)
    plt.setActiveAxes(-1)
    plt.update()
    return ax


def save(path, figsize=None, dpi=None):
    """Save plot.

    Keyword arguments:
    path -- Destination file path.
    figsize -- w,h figure size tuple in inches.
    dpi -- Dots per inch.
    """
    plt = getPlot()
    if not plt:
        return
    # Backup figure options
    fig = plt.fig
    sizeBack = fig.get_size_inches()
    dpiBack = fig.get_dpi()
    # Save figure with new options
    if figsize:
        fig.set_size_inches(figsize[0], figsize[1])
    if dpi:
        fig.set_dpi(dpi)
    plt.canvas.print_figure(path)
    # Restore figure options
    fig.set_size_inches(sizeBack[0], sizeBack[1])
    fig.set_dpi(dpiBack)
    plt.update()


def addNavigationToolbar():
    """Add the matplotlib QT navigation toolbar to the plot."""
    plt = getPlot()
    if not plt:
        return
    # Check that the navigation toolbar has not been already created
    if plt.mpl_toolbar is not None:
        return
    # Create the navigation toolbar and add it
    plt.mpl_toolbar = NavigationToolbar(plt.canvas, plt)
    vbox = plt.layout()
    vbox.addWidget(plt.mpl_toolbar)


def delNavigationToolbar():
    """Remove the matplotlib QT navigation toolbar from the plot."""
    plt = getPlot()
    if not plt:
        return
    # Check that the navigation toolbar already exist
    if plt.mpl_toolbar is None:
        return
    # Remove the widget from the layout
    vbox = plt.layout()
    vbox.removeWidget(plt.mpl_toolbar)
    # Destroy the navigation toolbar
    plt.mpl_toolbar.deleteLater()
    plt.mpl_toolbar = None


class Line:
    def __init__(self, axes, x, y, name):
        """Construct a new plot serie.

        Keyword arguments:
        axes -- Active axes
        x -- X values
        y -- Y values
        name -- Data serie name (for legend).
        """
        self.axes = axes
        self.x = x
        self.y = y
        self.name = name
        self.lid = len(axes.lines)
        (self.line,) = axes.plot(x, y)

    def setp(self, prop, value):
        """Change a line property value.

        Keyword arguments:
        prop -- Property name.
        value -- New property value.
        """
        plt.setp(self.line, prop, value)

    def getp(self, prop):
        """Get line property value.

        Keyword arguments:
        prop -- Property name.
        """
        return plt.getp(self.line, prop)


class Plot(PySide.QtGui.QWidget):
    def __init__(self, winTitle="plot", parent=None, flags=PySide.QtCore.Qt.WindowFlags(0)):
        """Construct a new plot widget.

        Keyword arguments:
        winTitle -- Tab title.
        parent -- Widget parent.
        flags -- QWidget flags
        """
        PySide.QtGui.QWidget.__init__(self, parent, flags)
        self.setWindowTitle(winTitle)
        # Create matplotlib canvas
        self.fig = Figure()
        self.canvas = FigureCanvas(self.fig)
        self.canvas.setParent(self)
        # Get axes
        self.axes = self.fig.add_subplot(111)
        self.axesList = [self.axes]
        self.axes.xaxis.set_ticks_position("bottom")
        self.axes.spines["top"].set_color("none")
        self.axes.yaxis.set_ticks_position("left")
        self.axes.spines["right"].set_color("none")
        # Add the navigation toolbar by default
        self.mpl_toolbar = NavigationToolbar(self.canvas, self)
        # Setup layout
        vbox = PySide.QtGui.QVBoxLayout()
        vbox.addWidget(self.canvas)
        vbox.addWidget(self.mpl_toolbar)
        self.setLayout(vbox)
        # Active series
        self.series = []
        # Indicators
        self.skip = False
        self.legend = False
        self.legPos = (1.0, 1.0)
        self.legSiz = 14
        self.grid = False

    def plot(self, x, y, name=None):
        """Plot a new line and return it.

        Keyword arguments:
        x -- X values
        y -- Y values
        name -- Serie name (for legend)."""
        l = Line(self.axes, x, y, name)
        self.series.append(l)
        # Update window
        self.update()
        return l

    def update(self):
        """Update the plot, redrawing the canvas."""
        if not self.skip:
            self.skip = True
            if self.legend:
                legend(self.legend, self.legPos, self.legSiz)
            self.canvas.draw()
            self.skip = False

    def isGrid(self):
        """Return True if Grid is active, False otherwise."""
        return bool(self.grid)

    def isLegend(self):
        """Return True if Legend is active, False otherwise."""
        return bool(self.legend)

    def setActiveAxes(self, index):
        """Change the current active axes.

        Keyword arguments:
        index -- Index of the new active axes set.
        """
        self.axes = self.axesList[index]
        self.fig.sca(self.axes)





import FreeCAD
import os
import unittest
from .TechDrawTestUtilities import createPageWithSVGTemplate


class DrawViewImageTest(unittest.TestCase):
    def setUp(self):
        """Creates a page"""
        FreeCAD.newDocument("TDAnno")
        FreeCAD.setActiveDocument("TDAnno")
        FreeCAD.ActiveDocument = FreeCAD.getDocument("TDAnno")
        self.page = createPageWithSVGTemplate()

    def tearDown(self):
        FreeCAD.closeDocument("TDAnno")

    def testMakeImage(self):
        """Tests if an image can be added to page"""
        path = os.path.dirname(os.path.abspath(__file__))
        imageFileSpec = path + "/TestImage.png"
        img = FreeCAD.ActiveDocument.addObject("TechDraw::DrawViewImage", "TestImage")
        img.ImageFile = imageFileSpec
        self.page.addView(img)

        FreeCAD.ActiveDocument.recompute()

        self.assertTrue("Up-to-date" in img.State)


if __name__ == "__main__":
    unittest.main()
